/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: BIU.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: Bus Interface Unit
//  
// Architecture Notes:
// 	Connect CPU with external worlds, such as bridge and memory controller. 
// 	Internally, it have to arbitrate which of instruction and data requests 
//  	to be serviced.
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$
`include "pool/timescale.v"
`include "pool/sys.h"

// BIU: bus interface unit
module BIU(
	INTREQN,		// interrupt request low active

// IU interface
	iALL_UPDATE_IFA_W,
        INU_ADDR,
        INU_DATA,
	INU_UC,			// insn in uncached region
        INU_REQ,

        iBIU2INU_BUSY,		// BIU tells INU that she's busy, or bridge is busy

	iLSU_ADDR_MR,
        LSU_ADDR,
        LSU_DI,			// LSU's data input
        LSU_DO,			
        LSU_STORREQ,
        LSU_SIZE,
        LSU_LOADREQ,
	LSU_UC,			// LSU uncached region

        iBIU2LSU_BUSY,		// BIU tells LSU that she's busy, or bridge is busy 

// OR1K native bus interface(NBI)
	NBI_ADDR,		// addr out
	NBI_DO,			// data out
	NBI_SIZE,		// 2 bit size
	NBI_RW,			// read(1)/write(0)
	NBI_BURSTN,		// burst transaction
	NBI_REQN,		// NBI request to bridge

	NBI_DI,			// data in
	NBI_GNTN,		// bridge grants the request of NBI_REQN

// system
        RSTN,
        CLK
);
	input  [31:0]	iLSU_ADDR_MR;

	output [31:0]	NBI_ADDR;
	output [31:0]	NBI_DO;
	output [1:0]	NBI_SIZE;
	output		NBI_RW;
	output		NBI_BURSTN;
	output		NBI_REQN;
	
	input  [31:0]	NBI_DI;
	input		NBI_GNTN;	

	input  [2:0]	INTREQN;

	input		iALL_UPDATE_IFA_W;
	input  [31:0]	INU_ADDR;
	output [31:0]	INU_DATA;
	input		INU_REQ;
	input		INU_UC;
	output		iBIU2INU_BUSY;

	output		iBIU2LSU_BUSY;
	input  [31:0]	LSU_ADDR;
	output [31:0]	LSU_DI;
	input  [31:0]	LSU_DO;
	input		LSU_STORREQ;
	input  [1:0]	LSU_SIZE;
	input  		LSU_LOADREQ;
	input		LSU_UC;


	input		RSTN, CLK;

// internal signals/registers
	reg    [1:0]	NBI_SIZE, iNBI_SIZE;	
	reg    [31:0]	NBI_ADDR, iNBI_ADDR;
	reg		NBI_RW, iNBI_RW;
	reg		NBI_BURSTN, iNBI_BURSTN;
	reg		NBI_REQN, iNBI_REQN;
	reg    [31:0]	NBI_DO, iNBI_DO;
	reg    [1:0]	NBI_CNT, iNBI_CNT;
	reg		BIU2INU_BUSY, iBIU2INU_BUSY;
	reg		BIU2LSU_BUSY, iBIU2LSU_BUSY;
	reg    [31:0]	LSU_DI, iLSU_DI;
	reg    [31:0]	INU_DATA, iINU_DATA;
	reg    [3:0]	NBISM, iNBISM;
	reg		ALL_UPDATE_IFA_WR, iALL_UPDATE_IFA_WR;
	reg		INU_REQ_D1;
	reg		SLAVE_IGNT, iSLAVE_IGNT;
	reg		SLAVE_DGNT, iSLAVE_DGNT;
	reg		SIMUL_ID_BUSY, iSIMUL_ID_BUSY;

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0) begin
		NBI_SIZE   	<= #(`UD) 2'b11;
		NBI_ADDR   	<= #(`UD) 32'h0000_0100;
		NBI_RW     	<= #(`UD) 1'b1;
		NBI_BURSTN 	<= #(`UD) 1'b1;
		NBI_REQN   	<= #(`UD) 1'b1;
		NBI_DO     	<= #(`UD) 32'hxxxx_xxxx;


		BIU2INU_BUSY	<= #(`UD) 1'b0;
		BIU2LSU_BUSY	<= #(`UD) 1'b0;

		NBI_CNT		<= #(`UD) 2'b00;
		NBISM		<= #(`UD) `NBISM_IDLE;

		LSU_DI		<= #(`UD) 32'hxxxx_xxxx;

		INU_REQ_D1	<= #(`UD) 1'b0;
		ALL_UPDATE_IFA_WR	<= #(`UD) 1'b0;
		SLAVE_IGNT	<= #(`UD) 1'b0;
		SLAVE_DGNT	<= #(`UD) 1'b0;
		SIMUL_ID_BUSY	<= #(`UD) 1'b0;
	end
	else begin
		NBI_SIZE   	<= #(`UD) iNBI_SIZE;
		NBI_ADDR   	<= #(`UD) iNBI_ADDR;
		NBI_RW     	<= #(`UD) iNBI_RW;
		NBI_BURSTN 	<= #(`UD) iNBI_BURSTN;
		NBI_REQN   	<= #(`UD) iNBI_REQN;
		NBI_DO     	<= #(`UD) iNBI_DO;

		BIU2INU_BUSY	<= #(`UD) iBIU2INU_BUSY;
		BIU2LSU_BUSY	<= #(`UD) iBIU2LSU_BUSY;

		NBI_CNT		<= #(`UD) iNBI_CNT;
		NBISM		<= #(`UD) iNBISM;
		
		LSU_DI		<= #(`UD) iLSU_DI;
		INU_DATA	<= #(`UD) iINU_DATA;

		INU_REQ_D1	<= #(`UD) INU_REQ;
		ALL_UPDATE_IFA_WR	<= #(`UD) iALL_UPDATE_IFA_WR;	
		SLAVE_IGNT	<= #(`UD) iSLAVE_IGNT;
		SLAVE_DGNT	<= #(`UD) iSLAVE_DGNT;
		SIMUL_ID_BUSY	<= #(`UD) iSIMUL_ID_BUSY;
	end

always @(NBI_GNTN or BIU2INU_BUSY or 
	 INU_REQ or LSU_LOADREQ or LSU_STORREQ or BIU2LSU_BUSY or
	 NBISM or iNBISM or
	 INU_REQ or INU_REQ_D1 or iALL_UPDATE_IFA_W or
	 ALL_UPDATE_IFA_WR or SLAVE_IGNT or SLAVE_DGNT
) begin
	if(iALL_UPDATE_IFA_W)
		iALL_UPDATE_IFA_WR = 1'b1;
	else if(NBISM == `NBISM_IRACCLAST && ~NBI_GNTN)
		iALL_UPDATE_IFA_WR = 1'b0;
	else
		iALL_UPDATE_IFA_WR = ALL_UPDATE_IFA_WR;

	if(NBISM == `NBISM_IRACCLAST && ~NBI_GNTN && ~iALL_UPDATE_IFA_W || SLAVE_IGNT)
		iBIU2INU_BUSY = 1'b0;

	else if(NBISM == `NBISM_IRACCLAST && ~NBI_GNTN && iALL_UPDATE_IFA_W)
		iBIU2INU_BUSY = 1'b1;

	else if(ALL_UPDATE_IFA_WR && iALL_UPDATE_IFA_W && ~BIU2LSU_BUSY || INU_REQ && ~INU_REQ_D1)
		iBIU2INU_BUSY = 1'b1;
	else
		iBIU2INU_BUSY = BIU2INU_BUSY;

	if(NBISM == `NBISM_IRACCLAST && ~NBI_GNTN && iALL_UPDATE_IFA_W)
		iSLAVE_IGNT = 1'b1;
	else if(SLAVE_IGNT && ~iBIU2INU_BUSY)
		iSLAVE_IGNT = 1'b0;
	else
		iSLAVE_IGNT = SLAVE_IGNT;

	if((NBISM == `NBISM_DRACCLAST || NBISM == `NBISM_DWACCLAST) && ~NBI_REQN && ~NBI_GNTN)
		iSLAVE_DGNT = 1'b1;
	else if(SLAVE_DGNT && ~iBIU2LSU_BUSY)
		iSLAVE_DGNT = 1'b0;
	else
		iSLAVE_DGNT = SLAVE_DGNT;

	if(~NBI_GNTN && NBISM == `NBISM_DWACCLAST)
		iBIU2LSU_BUSY = 1'b0;
	else if(~NBI_GNTN && NBISM == `NBISM_DRACCLAST)
		iBIU2LSU_BUSY = 1'b0;

	else if((LSU_LOADREQ || LSU_STORREQ) && (~BIU2INU_BUSY || SLAVE_IGNT) && 
		 ~(SLAVE_DGNT && iBIU2INU_BUSY))

		iBIU2LSU_BUSY = 1'b1;
	else
		iBIU2LSU_BUSY = BIU2LSU_BUSY;
end

always @(LSU_LOADREQ or LSU_STORREQ or
	 LSU_UC or LSU_DO or LSU_ADDR or LSU_SIZE or iLSU_ADDR_MR or
	 INU_ADDR or INU_UC or INU_REQ or
	 NBI_SIZE or NBI_ADDR or NBI_RW or NBI_BURSTN or
	 NBI_CNT or NBI_REQN or NBI_GNTN or
	 NBI_DI or INU_DATA or LSU_DI or
	 NBISM or iALL_UPDATE_IFA_WR or
	 INU_REQ_D1 or NBI_DO or 
	 iBIU2LSU_BUSY or iBIU2INU_BUSY or
	 BIU2LSU_BUSY or BIU2INU_BUSY or SIMUL_ID_BUSY or
	 SLAVE_DGNT
) begin
	if(NBISM == `NBISM_IRACCLAST && ~NBI_REQN && ~NBI_GNTN) // insn access
		iINU_DATA = NBI_DI;
	else
		iINU_DATA = INU_DATA;

	if(NBISM == `NBISM_DRACCLAST && ~NBI_REQN && ~NBI_GNTN) // insn access
		iLSU_DI = NBI_DI;
	else
		iLSU_DI = LSU_DI;

	if(iBIU2LSU_BUSY && iBIU2INU_BUSY && ~BIU2LSU_BUSY && ~BIU2INU_BUSY)
		iSIMUL_ID_BUSY = 1'b1;
	else if(~NBI_REQN && ~NBI_GNTN)
		iSIMUL_ID_BUSY = 1'b0;
	else
		iSIMUL_ID_BUSY = SIMUL_ID_BUSY;
		
		
	if(NBISM == `NBISM_IDLE && ((LSU_LOADREQ || LSU_STORREQ) && iBIU2LSU_BUSY)) begin
		iNBI_SIZE   = LSU_SIZE;
		iNBI_ADDR   = iLSU_ADDR_MR;
		iNBI_RW     = LSU_LOADREQ;
		iNBI_BURSTN = LSU_UC | ~LSU_UC & LSU_LOADREQ;
		iNBI_DO     = LSU_DO;
	end
	else if(NBISM == `NBISM_IDLE && iALL_UPDATE_IFA_WR) begin
		iNBI_SIZE   = 2'b11;
		iNBI_ADDR   = INU_ADDR;
		iNBI_RW     = 1'b1;
		iNBI_BURSTN = INU_UC;
		iNBI_DO     = 32'hxxxx_xxxx;
	end
	else begin
		iNBI_SIZE   = NBI_SIZE;
		iNBI_ADDR   = NBI_ADDR;
		iNBI_RW     = NBI_RW;
		iNBI_BURSTN = NBI_BURSTN;
		iNBI_DO     = NBI_DO;
	end

	if(~NBI_BURSTN && ~NBI_GNTN)
		iNBI_CNT = NBI_CNT + 1'b1;
	else
		iNBI_CNT = 2'b00;

	if(NBISM == `NBISM_IDLE && 
	   (LSU_LOADREQ || LSU_STORREQ || iALL_UPDATE_IFA_WR || INU_REQ && ~INU_REQ_D1))
		iNBI_REQN = 1'b0;
	else if(~NBI_GNTN && (NBI_BURSTN || ~NBI_BURSTN && NBI_CNT == 2'b11))
		iNBI_REQN = 1'b1;
	else
		iNBI_REQN = NBI_REQN;

end

// NBISM: native bus interface state machine
always @(NBISM or LSU_LOADREQ or LSU_STORREQ or iALL_UPDATE_IFA_WR or
	 NBI_GNTN or INU_REQ or INU_REQ_D1 or iBIU2LSU_BUSY
) begin
	case(NBISM) // synopsys parallel_case
	`NBISM_IDLE: begin
			if(LSU_LOADREQ && iBIU2LSU_BUSY)
			  iNBISM = `NBISM_DRACC;
			else if(LSU_STORREQ && iBIU2LSU_BUSY)
			  iNBISM = `NBISM_DWACC;
			else if(iALL_UPDATE_IFA_WR || INU_REQ && ~INU_REQ_D1)
			  iNBISM = `NBISM_IRACC;
			else
			  iNBISM = `NBISM_IDLE;
		     end
	`NBISM_IRACC:begin
			iNBISM = `NBISM_IRACCLAST;
		     end
	`NBISM_IRACCLAST:
	             begin
			if(~NBI_GNTN)
			  iNBISM = `NBISM_IDLE;
			else
			  iNBISM = `NBISM_IRACCLAST;
		     end
	`NBISM_DRACC:begin
			iNBISM = `NBISM_DRACCLAST;
		     end
	`NBISM_DRACCLAST:
		     begin
			if(~NBI_GNTN)
			  iNBISM = `NBISM_IDLE;
			else
			  iNBISM = `NBISM_DRACCLAST;
		     end
	`NBISM_DWACC:begin
			iNBISM = `NBISM_DWACCLAST;
		     end
	`NBISM_DWACCLAST:  
		     begin
			if(~NBI_GNTN)		
			  iNBISM = `NBISM_IDLE;
		        else
			  iNBISM = `NBISM_DWACCLAST;
		     end
	default:     begin iNBISM = `NBISM_IDLE; $display("NBISM enters wrong state"); end
	endcase
end

endmodule
