/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: BRIDGE.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: test bench
//  
// Architecture Notes:
// 	auto load next program to be tested if current one PASSes. 
//  
//  
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$
`include "pool/timescale.v"
`include "pool/sys.h"

module BRIDGE(
        NBI_ADDR,               // addr out
        NBI_DO,                 // data out
        NBI_SIZE,               // 2 bit size
        NBI_RW,                 // read(1)/write(0)
        NBI_BURSTN,             // burst transaction
        NBI_REQN,               // NBI request to bridge

        NBI_DI,                 // data in
        NBI_GNTN,               // bridge grants the request of NBI_REQN

// system
        RSTN,
        CLK
);

	parameter ABITS = 32; 
	parameter DBITS = 32;

	input  [ABITS-1:0] NBI_ADDR;
	output [DBITS-1:0] NBI_DI;
	input  [DBITS-1:0] NBI_DO;
	input		   NBI_RW;
	input		   NBI_BURSTN;
	input		   NBI_REQN;
	output	   	   NBI_GNTN;
	input  [1:0]	   NBI_SIZE;
	input		   RSTN, CLK;
	
	reg    [DBITS-1:0] MPOOL[0:(1<<16)-1]; // 64 K words/256K bytes
	reg    [31:0]	NBI_DI;
	reg    [1:0]	BURST_CNT;
	reg		NBI_GNTN;
	reg    [31:0]	NBI_ADDR_PLUS4, iNBI_ADDR_PLUS4;
	reg    [31:0]	DOTMP;

	wire [31:0] BRIDGE_DO = MPOOL[iNBI_ADDR_PLUS4[31:2]];

	wire [31:2] NBI_ADDR31_2 = NBI_ADDR[31:2];

	always @(negedge CLK)
	if(~NBI_REQN && ~NBI_GNTN && ~NBI_RW && NBI_BURSTN) begin // NBI write

		DOTMP = MPOOL[NBI_ADDR31_2];

		case({NBI_SIZE, NBI_ADDR[1:0]})
		4'b00_00: MPOOL[NBI_ADDR[31:2]] = {NBI_DO[31:24], DOTMP[23:0]};
		4'b00_01: MPOOL[NBI_ADDR[31:2]] = {DOTMP[31:24], NBI_DO[23:16], DOTMP[15:0]};
		4'b00_10: MPOOL[NBI_ADDR[31:2]] = {DOTMP[31:16], NBI_DO[15:8], DOTMP[7:0]};
		4'b00_11: MPOOL[NBI_ADDR[31:2]] = {DOTMP[31:8], NBI_DO[7:0]};
		4'b01_00: MPOOL[NBI_ADDR[31:2]] = {NBI_DO[31:16], DOTMP[15:0]};
		4'b01_10: MPOOL[NBI_ADDR[31:2]] = {DOTMP[31:16], NBI_DO[15:0]};
		4'b11_00: MPOOL[NBI_ADDR[31:2]] = NBI_DO;
		default:  begin $display("NBI write wrong address and size"); #100 $finish; end
		endcase
	end

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0)
		NBI_DI <= #(`UD) 32'hxxxx_xxxx;

	else if(~NBI_REQN && NBI_RW) begin // NBI read
		if(~NBI_BURSTN) begin
			NBI_DI <= #(`UD) BRIDGE_DO;
		end
		else begin
		  case({NBI_SIZE, NBI_ADDR[1:0]})
		  4'b00_00: NBI_DI <= #(`UD) {BRIDGE_DO[31:24], 24'hxx_xxxx};
		  4'b00_01: NBI_DI <= #(`UD) {8'hxx, BRIDGE_DO[23:16], 16'hxxxx};
		  4'b00_10: NBI_DI <= #(`UD) {16'hxx, BRIDGE_DO[15:8], 8'hxx};
		  4'b00_11: NBI_DI <= #(`UD) {24'hxx, BRIDGE_DO[7:0]};
		  4'b01_00: NBI_DI <= #(`UD) {BRIDGE_DO[31:16], 16'hxxxx};
		  4'b01_10: NBI_DI <= #(`UD) {16'hxxxx, BRIDGE_DO[15:0]};
		  4'b11_00: NBI_DI <= #(`UD) BRIDGE_DO;
		  default:  begin $display("PA=%x", NBI_ADDR);
				  $display("NBI read wrong address and size at time=%d", $time); #100 $finish; end
		  endcase
		end
	end

reg [7:0] 	GNTN_CNT, RANDOM;

`define GNTN_END 1

  `define RANDOM_GNTN_END
  `ifdef RANDOM_GNTN_END
        initial begin
		RANDOM = `GNTN_END;
	end
  `else
	always @(posedge CLK or negedge RSTN)
		if(RSTN == 1'b0)
			RANDOM <= #8'b1;
		else if(~NBI_REQN && ~NBI_GNTN)
			RANDOM <= #(`UD) RANDOM + 1'b1;
  `endif

	always @(posedge CLK or negedge RSTN)
		if(RSTN == 1'b0)
			GNTN_CNT <= 8'b0;
		else if(~NBI_REQN && NBI_BURSTN && GNTN_CNT == RANDOM[4:0])
			GNTN_CNT <= #(`UD) 8'b0;
		else if(~NBI_REQN && NBI_BURSTN && GNTN_CNT != RANDOM[4:0])
			GNTN_CNT <= #(`UD) GNTN_CNT + 1'b1;

	always @(posedge CLK or negedge RSTN)
		if(RSTN == 1'b0)
			NBI_GNTN <= #(`UD) 1'b1;
		else if(~NBI_REQN && NBI_BURSTN && 
			(GNTN_CNT == RANDOM[4:0] || GNTN_CNT == 8'b0 && ~NBI_GNTN)) begin
			NBI_GNTN <= #(`UD) ~NBI_GNTN;
		end
		else if(~NBI_REQN && ~NBI_BURSTN) begin
			if(BURST_CNT == 2'b00)
				NBI_GNTN <= #(`UD) ~NBI_GNTN;
		end

	always @(posedge CLK or negedge RSTN)
		if(RSTN == 1'b0)
			NBI_ADDR_PLUS4 <= #(`UD) 32'h0000_0100;
		else if(~NBI_REQN && ~NBI_BURSTN)
			NBI_ADDR_PLUS4 <= #(`UD) iNBI_ADDR_PLUS4;
	
always @(NBI_REQN or NBI_BURSTN or BURST_CNT or NBI_ADDR or NBI_ADDR_PLUS4) begin

	if(~NBI_REQN && ~NBI_BURSTN) begin
	  if(BURST_CNT == 2'b00)
	    iNBI_ADDR_PLUS4 = NBI_ADDR[31:0];
	  else
	    iNBI_ADDR_PLUS4 = {NBI_ADDR[31:4], NBI_ADDR_PLUS4[3:2] + 1, NBI_ADDR[1:0]};
	end
	else if(~NBI_REQN && NBI_BURSTN) begin
		iNBI_ADDR_PLUS4 = NBI_ADDR[31:0];
	end

end

	always @(posedge CLK or negedge RSTN)
		if(RSTN == 1'b0)
			BURST_CNT <= #(`UD) 2'b00;
		else if(~NBI_REQN && ~NBI_BURSTN)
			BURST_CNT <= #(`UD) BURST_CNT + 1'b1;
		else
			BURST_CNT <= #(`UD) 2'b00;
			


wire [31:0] d0  = MPOOL[12'h100>>2];
wire [31:0] d1  = MPOOL[12'h104>>2];
wire [31:0] d2  = MPOOL[12'h108>>2];

`ifdef  TRANSLATE_ON  //* debug all programs
	reg  [11:0]	PGM_PTR; //* pointer to sequential programs
	reg  [8*40-1:0] FILENAME[0:127]; //* filename length 40 characters
	reg  [8*40-1:0] cur_filename; //* current file name
	integer		i;

	initial begin
		i = 0;
`define PGM_STX 17
		FILENAME[0]  = "pgm/obj/addsubs1a.hex";
		FILENAME[1]  = "pgm/obj/addsubs1b.hex";
		FILENAME[2]  = "pgm/obj/addsubu1a.hex";
		FILENAME[3]  = "pgm/obj/addsubu1b.hex";
		FILENAME[4]  = "pgm/obj/bxf1a.hex";
		FILENAME[5]  = "pgm/obj/bxf1b.hex";
		FILENAME[6]  = "pgm/obj/jjal1a.hex";
		FILENAME[7]  = "pgm/obj/jjal1b.hex";
		FILENAME[8]  = "pgm/obj/ldst1a.hex";
		FILENAME[9]  = "pgm/obj/logic1a.hex";
		FILENAME[10] = "pgm/obj/logic1b.hex";
		FILENAME[11] = "pgm/obj/bit1a.hex";
		FILENAME[12] = "pgm/obj/bit1b.hex";
		FILENAME[13] = "pgm/obj/shift1a.hex";
		FILENAME[14] = "pgm/obj/swap1a.hex";
		FILENAME[15] = "pgm/obj/jjaljr1a.hex";
		FILENAME[16] = "pgm/obj/jjaljr1b.hex";
		FILENAME[17] = "pgm/obj/jjalr1a.hex";
		FILENAME[18] = "pgm/obj/jjalr1b.hex";
		FILENAME[19] = "pgm/obj/ext1a.hex";
		FILENAME[20] = "pgm/obj/ldst1b.hex";
		FILENAME[21] = "pgm/obj/addsubwcs1a.hex";
		FILENAME[22] = "pgm/obj/sysbrk1a.hex";
		FILENAME[23] = "pgm/obj/sysbrk2a.hex";
		FILENAME[24] = "pgm/obj/overfl1a.hex";
		FILENAME[25] = "pgm/obj/overfl2a.hex";
		FILENAME[26] = "pgm/obj/aae1a.hex";
		FILENAME[27] = "pgm/obj/aae2a.hex";
		FILENAME[28] = "pgm/obj/mtmfsr1a.hex";
		FILENAME[29] = "pgm/obj/int1a.hex";
                FILENAME[30] = "pgm/obj/ldst1c.hex";
                FILENAME[31] = "pgm/obj/ldstpai1a.hex";
                FILENAME[32] = "pgm/obj/ldstpai2a.hex";
                FILENAME[33] = "pgm/obj/ldstpai2b.hex";
                FILENAME[34] = "pgm/obj/ldstpai2c.hex";
                FILENAME[35] = "pgm/obj/ldstpai3a.hex";
                FILENAME[36] = "pgm/obj/ldstpai3b.hex";
                FILENAME[37] = "pgm/obj/ldstpai4a.hex";
                FILENAME[38] = "pgm/obj/ldstpai4b.hex";
                FILENAME[39] = "pgm/obj/ldstpai4c.hex";
                FILENAME[40] = "pgm/obj/ldstpai5a.hex";
                FILENAME[41] = "pgm/obj/ldstpai5b.hex";
                FILENAME[42] = "pgm/obj/ldstpai5c.hex";
                FILENAME[43] = "pgm/obj/ldstpai5d.hex";
                FILENAME[44] = "pgm/obj/ldstpai6a.hex";
		FILENAME[45] = "pgm/obj/iie1a.hex";

	end

  `define RSTN top.RSTN


  //`define DECREMENT
  `ifdef DECREMENT
	initial PGM_PTR  = `PGM_STX;
  `else
	initial PGM_PTR  = 0;
  `endif

	always @(posedge `RSTN) begin
		for(i=0; i<(1<<ABITS)-1; i=i+1)
			MPOOL[i] = 32'hxxxx_xxxx;

		load_file("pgm/obj/RESET.hex",  (20'h00100>>2));
		load_file("pgm/obj/SYS.hex",    (20'h00C00>>2));
		load_file("pgm/obj/BRK.hex",    (20'h00D00>>2));

		load_file("pgm/obj/AAE.hex",	(20'h00600>>2)); // addr alignment exception
		load_file("pgm/obj/IIE.hex",	(20'h00700>>2)); // illegal insn exception
		load_file("pgm/obj/INT.hex",    (20'h00800>>2));
		load_file("pgm/obj/OVERFL.hex", (20'h00E00>>2));
		//load_file("pgm/obj/DIVZERO.hex",(20'h00F00>>2));

		load_file(FILENAME[PGM_PTR], (20'h20000>>2));

		cur_filename 	<= #(`UD) FILENAME[PGM_PTR];
  `ifdef DECREMENT
		PGM_PTR 	<= #(`UD) PGM_PTR - 1'b1;
  `else
		PGM_PTR 	<= #(`UD) PGM_PTR + 1'b1;
  `endif
	end
`endif

task load_file;
	input [8*40-1:0] FILE;
	input [17:0] SA; // start address
	begin
	  $readmemh(FILE, MPOOL, SA); //* reset start address
	end
endtask

endmodule
