/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: DPCON.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: DataPath CONtrol logic
//  
// Architecture Notes:
// 	It controls the datapath forwarding logic(DPFOR).
//  
//  
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$
`include "pool/timescale.v"
`include "pool/sys.h"
`include "pool/or1601isa.h"

module DPCON(
	STALL_DPCON,

	iAOPCON,
	iBOPCON,
	iSTORCON,

	iIALU_IMME_CONST,
	iSTOR_IMME_CONST,

        iRF_AADR,
        iRF_BADR,
        iRF_AREN,
        iRF_BREN,
        iRF_WADR_M,
        iRF_WEN_M,

        RF_WADR_ER,
        RF_WADR_MR,
        RF_WADR_WR,
        RF_WEN_ER,
        RF_WEN_MR,
        RF_WEN_WR,
	
        iSEL_RF_WDI_WR,

	
	JAL_JALR_MR,
	LSU_LOADREQ_MR,
	LSU_LOADREQ,
	STORPAI_MR,			// for h.swpai/h.shpai/h.sbpai

	MFSR_ER,
	MFSR_MR,

	CLK,
	RSTN
);
	input		STORPAI_MR;
	input		JAL_JALR_MR;
	input		LSU_LOADREQ, LSU_LOADREQ_MR, MFSR_ER, MFSR_MR;
	output [3:0]	iSEL_RF_WDI_WR;

	output [7:0]	iAOPCON;
	output [7:0]	iBOPCON;
	output [7:0]	iSTORCON;

	input		iIALU_IMME_CONST;
	input		iSTOR_IMME_CONST;

	input  [3:0]	iRF_AADR, iRF_BADR;
	input		iRF_AREN, iRF_BREN;
	input  [3:0]	iRF_WADR_M;
	input		iRF_WEN_M;

	input  [3:0]	RF_WADR_ER, RF_WADR_MR, RF_WADR_WR;
	input       	RF_WEN_ER, RF_WEN_MR, RF_WEN_WR;
	
	input		STALL_DPCON;
	input		CLK, RSTN;


	reg		RF_BREN_ER;
	reg    [3:0]	iSEL_RF_WDI_WR;
	reg    [7:0]	iAOPCON, iBOPCON;
	reg		iA_FOR_E, iA_FOR_M, iA_FOR_W;
	reg		iB_FOR_E, iB_FOR_M, iB_FOR_W;
	reg		iSTOR_FOR_M, iSTOR_FOR_W;
	reg    [3:0]	RF_BADR_ER;
	reg    [7:0]	iSTORCON;
	reg		iLSU_WRITEBACK_EN;

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0) begin
		RF_BREN_ER 	<= #(`UD) 1'b0;
		RF_BADR_ER	<= #(`UD) 4'b0;
	end
	else begin
		RF_BREN_ER	<= #(`UD) (STALL_DPCON) ? RF_BREN_ER : iRF_BREN;
		RF_BADR_ER	<= #(`UD) (STALL_DPCON) ? RF_BADR_ER : iRF_BADR;
	end

always @(STORPAI_MR or LSU_LOADREQ_MR
) begin
	// STORPAI_MR: for h.swpai, h.shpai, h.sbpai

	//iLSU_WRITEBACK_EN = (STORPAI_MR | LSU_LOADREQ_MR) ? 1'b1 : 1'b0;

	// <- we save the STORPAI adder by incrementing by 4/2/1 with IALU
	// <- use AOP_ER as the default address
	iLSU_WRITEBACK_EN = (LSU_LOADREQ_MR) ? 1'b1 : 1'b0;
end

always @(LSU_LOADREQ_MR or MFSR_MR or JAL_JALR_MR or iLSU_WRITEBACK_EN
) begin
	case({JAL_JALR_MR, MFSR_MR, iLSU_WRITEBACK_EN}) // synopsys parallel_case
	3'b100: iSEL_RF_WDI_WR = `SEL_RFWDI_SRA;
	3'b010: iSEL_RF_WDI_WR = `SEL_RFWDI_SPR;
	3'b001:	iSEL_RF_WDI_WR = `SEL_RFWDI_LSU;
	default:iSEL_RF_WDI_WR = `SEL_RFWDI_ALUMR;
	endcase
end

always @(iA_FOR_E or iA_FOR_M or iA_FOR_W or
         iB_FOR_E or iB_FOR_M or iB_FOR_W or iIALU_IMME_CONST or iSTOR_IMME_CONST or
	 iSTOR_FOR_M or iSTOR_FOR_W or LSU_LOADREQ or MFSR_ER or MFSR_MR
) begin
	//casex({LSU_LOADREQ, iA_FOR_E, iA_FOR_M, iA_FOR_W}) // synopsys parallel_case
	//casex({LSU_LOADREQ, (MFSR_ER|MFSR_MR), iA_FOR_E, iA_FOR_M, iA_FOR_W}) // synopsys parallel_case
	casex({LSU_LOADREQ, (MFSR_ER), iA_FOR_E, iA_FOR_M, iA_FOR_W}) // synopsys parallel_case
	5'b101xx: iAOPCON = `ASEL_LOAD_DI;
	5'b011xx: iAOPCON = `ASEL_SPR;
	5'b001xx: iAOPCON = `ASEL_IALU;
	5'bx001x: iAOPCON = `ASEL_RF_WDI_M;
	5'bx0001: iAOPCON = `ASEL_RF_WDI_WR;
	default:  iAOPCON = `ASEL_RF;
	endcase

	casex({LSU_LOADREQ, MFSR_ER, iSTOR_IMME_CONST, iIALU_IMME_CONST, iB_FOR_E, iB_FOR_M, iB_FOR_W}) // synopsys parallel_case
	7'bx01xxxx: iBOPCON = `BSEL_IMME_CNST;
	7'bx001xxx: iBOPCON = `BSEL_IMME_CNST;
	7'bx101xxx: iBOPCON = `BSEL_IMME_CNST;
	7'b00001xx: iBOPCON = `BSEL_IALU;
	7'b01001xx: iBOPCON = `BSEL_SPR;
	7'b10001xx: iBOPCON = `BSEL_LOAD_DI;
	7'bx00001x: iBOPCON = `BSEL_RF_WDI_M;
	7'bx000001: iBOPCON = `BSEL_RF_WDI_WR;
	default  : iBOPCON = `BSEL_RF;
	endcase

	casex({iSTOR_FOR_M, iSTOR_FOR_W})
	2'b1x:   iSTORCON = `STSEL_RF_WDI_M;
	2'b01:   iSTORCON = `STSEL_RF_WDI_WR;
	default: iSTORCON = `STSEL_RF;
	endcase
end


always @(iRF_AREN or 
	 RF_WEN_ER or RF_WEN_MR or RF_WEN_WR or
	 iRF_AADR or
	 RF_WADR_ER or RF_WADR_MR or RF_WADR_WR or
	 iRF_BREN or iRF_BADR
) begin
	if(iRF_AREN && ~RF_WEN_ER && (iRF_AADR == RF_WADR_ER))
		iA_FOR_E = 1'b1;
	else
		iA_FOR_E = 1'b0;

	if(iRF_AREN && ~RF_WEN_MR && (iRF_AADR == RF_WADR_MR))
		iA_FOR_M = 1'b1;
	else
		iA_FOR_M = 1'b0;

	if(iRF_AREN && ~RF_WEN_WR && (iRF_AADR == RF_WADR_WR))
		iA_FOR_W = 1'b1;
	else
		iA_FOR_W = 1'b0;

	if(iRF_BREN && ~RF_WEN_ER && (iRF_BADR == RF_WADR_ER))
		iB_FOR_E = 1'b1;
	else
		iB_FOR_E = 1'b0;

	if(iRF_BREN && ~RF_WEN_MR && (iRF_BADR == RF_WADR_MR))
		iB_FOR_M = 1'b1;
	else
		iB_FOR_M = 1'b0;

	if(iRF_BREN && ~RF_WEN_WR && (iRF_BADR == RF_WADR_WR))
		iB_FOR_W = 1'b1;
	else
		iB_FOR_W = 1'b0;
end

always @(RF_BREN_ER or RF_WEN_MR or RF_WEN_WR or
	 RF_BADR_ER or RF_WADR_MR or RF_WADR_WR
) begin
	if(RF_BREN_ER && ~RF_WEN_MR && RF_BADR_ER == RF_WADR_MR)
		iSTOR_FOR_M = 1'b1;
	else
		iSTOR_FOR_M = 1'b0;

	if(RF_BREN_ER && ~RF_WEN_WR && RF_BADR_ER == RF_WADR_WR)
		iSTOR_FOR_W = 1'b1;
	else
		iSTOR_FOR_W = 1'b0;
end

endmodule
