/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: DPFOR.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: DataPath FORwarding logic
//  
// Architecture Notes:
// 	It is the main data path module. 
//  
//  
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$
`include "pool/timescale.v"
`include "pool/sys.h"
`include "pool/or1601isa.h"

//* datapath bypassing/forwarding path
module DPFOR(
	PC_PLUS_4_MR,
	STALL_DPFOR,
	iMFSR_DELAY_SLOT_STALL,
	iLD_DELAY_SLOT_STALL,

	iAOP_ER,
	AOP_ER,		//* output to ALU
	BOP_ER,
	iSTOR_MR,	//* output to LSU
	STORPAI_MR,

	iAOPCON,
	iBOPCON,
	iSTORCON,

	RF_AO,		//* normal from D stage
	RF_BO,

	iIMME_CNST_D,	//* normal: const from D stage for B operand
	
	iALURES,	//* forward from E stage
	SPR_RES_M,	//* forward from E stage

	iLSU_DI2DP,		//* forward from M stage
	iRF_WDI_WR,	//* forward from M stage

	RF_WDI_WR,	//* forward from W stage

	iSEL_RF_WDI_WR,

	INU_DATA,

	CLK,
	RSTN
);
	input		iMFSR_DELAY_SLOT_STALL;
	input		iLD_DELAY_SLOT_STALL;

	input		STORPAI_MR;
	input  [31:0]	PC_PLUS_4_MR;
	input  [3:0]	iSEL_RF_WDI_WR;

	output [31:0]	iAOP_ER, AOP_ER, BOP_ER;

	output [31:0]	iSTOR_MR;
	input  [7:0]	iAOPCON, iBOPCON;
	input  [7:0]	iSTORCON;
	input  [31:0]	RF_AO, RF_BO;
	input  [31:0]	iIMME_CNST_D;
	input  [31:0]	iALURES, SPR_RES_M, iLSU_DI2DP;
	output [31:0]	iRF_WDI_WR, RF_WDI_WR;
	input		STALL_DPFOR;
	input  [31:0]	INU_DATA;
	input		CLK, RSTN;

	reg    [31:0]	AOP_ER, iAOP_ER;
	reg    [31:0]	BOP_ER, iBOP_ER;
	reg    [31:0]	STOR_MR, iSTOR_MR;
	reg    [31:0]	ALURES_MR, iALURES_MR;

	reg    [31:0]	RF_WDI_WR, iRF_WDI_WR;
	reg    [31:0]	RF_BO_ER;
	reg    [31:0]	iLSU_WDI_WR;

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0) begin
		AOP_ER		<= #(`UD) 32'h0000_0000;
		BOP_ER		<= #(`UD) 32'h0000_0000;
	end
	else begin
		AOP_ER		<= #(`UD) (STALL_DPFOR) ? AOP_ER : iAOP_ER;
		BOP_ER		<= #(`UD) (STALL_DPFOR) ? BOP_ER : iBOP_ER;

		ALURES_MR	<= #(`UD) (STALL_DPFOR) ? ALURES_MR : iALURES_MR;
		STOR_MR		<= #(`UD) (STALL_DPFOR) ? STOR_MR: iSTOR_MR;

		RF_BO_ER	<= #(`UD) (STALL_DPFOR) ? RF_BO_ER : RF_BO;

		//RF_WDI_WR	<= #(`UD) (STALL_DPFOR) ? RF_WDI_WR : iRF_WDI_WR;
		//RF_WDI_WR	<= #(`UD) (STALL_DPFOR & ~iMFSR_DELAY_SLOT_STALL) ? RF_WDI_WR : iRF_WDI_WR;
		RF_WDI_WR	<= #(`UD) (STALL_DPFOR & ~iMFSR_DELAY_SLOT_STALL & ~iLD_DELAY_SLOT_STALL) ? RF_WDI_WR : iRF_WDI_WR;
	end

always @(STORPAI_MR or iLSU_DI2DP
) begin
	//if(STORPAI_MR)
	//	iLSU_WDI_WR = LSU_ADDR_PLUS421_MR;
	//else

	// <- we save the STORPAI adder by incrementing by 4/2/1 with IALU
	// <- use AOP_ER as the default address

	iLSU_WDI_WR = iLSU_DI2DP; // DI(data in) to DP(data path)
end

always @(iALURES or iLSU_DI2DP or 
	 SPR_RES_M or iLSU_DI2DP or ALURES_MR or PC_PLUS_4_MR or
	 iSEL_RF_WDI_WR or iLSU_WDI_WR
) begin

	iALURES_MR = iALURES;

	case(iSEL_RF_WDI_WR) // synopsys parallel_case
	`SEL_RFWDI_SPR:   iRF_WDI_WR = SPR_RES_M;
	`SEL_RFWDI_LSU:   iRF_WDI_WR = iLSU_WDI_WR; 	// DI(data in) to DP(data path)
	`SEL_RFWDI_ALUMR: iRF_WDI_WR = ALURES_MR;
	`SEL_RFWDI_SRA:   iRF_WDI_WR = PC_PLUS_4_MR; 	// save return address
	default: 	  iRF_WDI_WR = 32'hxxxx_xxxx;
	endcase
end

always @(iAOPCON or RF_AO or iIMME_CNST_D or iALURES or SPR_RES_M or iRF_WDI_WR or RF_WDI_WR or
	 iLSU_DI2DP
) begin

	case(iAOPCON) // synopsys parallel_case
	`ASEL_LOAD_DI:  iAOP_ER = iLSU_DI2DP;		// there's no constant operand in AOP
	`ASEL_RF: 	iAOP_ER = RF_AO;		// constant operand all in BOP
	`ASEL_IALU: 	iAOP_ER = iALURES;
	`ASEL_SPR: 	iAOP_ER = SPR_RES_M;		// it's impossible here
	`ASEL_RF_WDI_M: iAOP_ER = iRF_WDI_WR;
	`ASEL_RF_WDI_WR:iAOP_ER = RF_WDI_WR;
	default:        iAOP_ER = 32'hxxxx_xxxx;	
	endcase
end

always @(iBOPCON or RF_AO or RF_BO or
	 iIMME_CNST_D or iALURES or SPR_RES_M or iRF_WDI_WR or RF_WDI_WR or INU_DATA or
	 iLSU_DI2DP
) begin
	case(iBOPCON) // synopsys parallel_case
	`BSEL_LOAD_DI:	iBOP_ER = iLSU_DI2DP;
	`BSEL_RF: 	iBOP_ER = RF_BO;
	`BSEL_IMME_CNST:iBOP_ER = iIMME_CNST_D;
	`BSEL_IALU: 	iBOP_ER = iALURES;
	`BSEL_SPR: 	iBOP_ER = SPR_RES_M;		// it's impossible here
	`BSEL_RF_WDI_M: iBOP_ER = iRF_WDI_WR;
	`BSEL_RF_WDI_WR:iBOP_ER = RF_WDI_WR;
	default:      	iBOP_ER = 32'hxxxx_xxxx;	
	endcase
end

always @(iSTORCON or RF_BO or iRF_WDI_WR or RF_WDI_WR or RF_BO_ER
) begin
	case(iSTORCON) // synopsys parallel_case
	`STSEL_RF:	 iSTOR_MR = RF_BO_ER;
	`STSEL_RF_WDI_M: iSTOR_MR = iRF_WDI_WR;
	`STSEL_RF_WDI_WR:iSTOR_MR = RF_WDI_WR;
	default:	 iSTOR_MR = 32'hxxxx_xxxx;
	endcase
end

endmodule
