/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: IALU.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: Integer ALU
//  
// Architecture Notes:
// 	Do all the ALU operations, except MUL/DIV/MOD/MAC. 
//  
//  
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$
`include "pool/timescale.v"
`include "pool/sys.h"
`include "pool/or1601isa.h"

//* integer ALU
module IALU(
	STALL_IALU,

//* operation code: the following signals should be the output of D-FF at E-stage
	SHOP_ER,
	SHLA_ER,
	SHRA_ER,
	SHRL_ER,
	ROR_ER,
	ROL_ER,

	SUBOP_ER,

	ADDOP_ER,
	ADDU_ER,
	SUBU_ER,
	ADDS_ER,
	SUBS_ER,
	ADDWCS_ER,
	SUBWCS_ER,	
	NEGS_ER,

	LOGOP_ER,
	XOR_ER,
	OR_ER,
	AND_ER,
	NOTU_ER,
	TEST_ER,
	SET_ER,
	CLR_ER,
	
	EXTOP_ER,
	EXTHS_ER,
	EXTHZ_ER,
	EXTBS_ER,
	EXTBZ_ER,
	
	MOVOP_ER,
	MOV_ER,
	MOVIBSE_ER,
	MOVIBZE_ER,
	MOVIBLS_ER,
	SWAPB_ER,
	SWAPH_ER,

//* operands
	CARRY_MR,
	CARRY_WR,

	AOP,
	BOP,
	iALURES,

	SFU_ER,
	SFS_ER,

	iFLAG_E,
	iCARRY_E,
	iOVERFL_E,

        MULDIVOP,
        MULDIV_RES,
        MULDIV_RES_ACK,

	RSTN,
	CLK
);
	input		STALL_IALU;

        input		SHLA_ER, SHRA_ER, SHRL_ER, ROR_ER, ROL_ER;
	input		ADDU_ER, SUBU_ER, ADDS_ER, SUBS_ER, ADDWCS_ER, SUBWCS_ER;
	input		XOR_ER, OR_ER, AND_ER, NOTU_ER, NEGS_ER, TEST_ER, CLR_ER, SET_ER;
	input		EXTHS_ER, EXTHZ_ER, EXTBS_ER, EXTBZ_ER;
	input		MOV_ER, MOVIBSE_ER, MOVIBZE_ER, MOVIBLS_ER, SWAPB_ER, SWAPH_ER;
	input		SHOP_ER, ADDOP_ER, LOGOP_ER, EXTOP_ER, MOVOP_ER;

	input		CARRY_MR, CARRY_WR;
	input  [31:0]	AOP;
	input  [31:0]	BOP;
	output [31:0]	iALURES;
	input  [3:0]	SUBOP_ER;
	input		SFU_ER, SFS_ER;
	output		iFLAG_E;
	output		iCARRY_E;
	output		iOVERFL_E;
	
	input		MULDIVOP;
	input  [31:0]	MULDIV_RES;
	input		MULDIV_RES_ACK;

	input		RSTN, CLK;

	reg    [31:0]	iSH_RES, tmp;
	reg    [31:0]	iADD_RES, iLOG_RES, iEXT_RES, iMOV_RES;
	reg    [31:0]	iALURES;
	reg		iFLAG_E, iOVERFL_E, iCARRY_E;
	reg		CARRY, iCARRY;
	reg		iALURES_SIGN, iALURES_ZERO;
	reg		tmp1bit;

always @(SUBOP_ER or SFU_ER or SFS_ER or iALURES_SIGN or iALURES_ZERO or
	 iALURES or TEST_ER or AOP or BOP
) begin
	if(iALURES == 32'b0)
		iALURES_ZERO = 1'b1;
	else
		iALURES_ZERO = 1'b0;

	iALURES_SIGN = iALURES[31];

	//case({SFU_ER, SFS_ER, SUBOP_ER, iALURES_SIGN, iALURES_ZERO})
	case({TEST_ER, SFU_ER, SFS_ER}) // synopsys parallel_case
	3'b100: begin
		  case(BOP[4:0])
		  0: iFLAG_E = AOP[0];
		  1: iFLAG_E = AOP[1];
		  2: iFLAG_E = AOP[2];
		  3: iFLAG_E = AOP[3];
		  4: iFLAG_E = AOP[4];
		  5: iFLAG_E = AOP[5];
		  6: iFLAG_E = AOP[6];
		  7: iFLAG_E = AOP[7];
		  8: iFLAG_E = AOP[8];
		  9: iFLAG_E = AOP[9];
		  10:iFLAG_E = AOP[10];
		  11:iFLAG_E = AOP[11];
		  12:iFLAG_E = AOP[12];
		  13:iFLAG_E = AOP[13];
		  14:iFLAG_E = AOP[14];
		  15:iFLAG_E = AOP[15];
		  16:iFLAG_E = AOP[16];
		  17:iFLAG_E = AOP[17];
		  18:iFLAG_E = AOP[18];
		  19:iFLAG_E = AOP[19];
		  20:iFLAG_E = AOP[20];
		  21:iFLAG_E = AOP[21];
		  22:iFLAG_E = AOP[22];
		  23:iFLAG_E = AOP[23];
		  24:iFLAG_E = AOP[24];
		  25:iFLAG_E = AOP[25];
		  26:iFLAG_E = AOP[26];
		  27:iFLAG_E = AOP[27];
		  28:iFLAG_E = AOP[28];
		  29:iFLAG_E = AOP[29];
		  30:iFLAG_E = AOP[30];
		  31:iFLAG_E = AOP[31];
		  default: iFLAG_E = 1'bx;
		  endcase
		end
	3'b010: begin
		case(SUBOP_ER) // synopsys parallel_case
		`H_SUBOP_SFGTU: iFLAG_E=(~iALURES_SIGN && ~iALURES_ZERO)?1'b1:1'b0;
		`H_SUBOP_SFGEU: iFLAG_E=(~iALURES_SIGN		 )?1'b1:1'b0;
		`H_SUBOP_SFLTU: iFLAG_E=( iALURES_SIGN && ~iALURES_ZERO)?1'b1:1'b0;
		`H_SUBOP_SFLEU: iFLAG_E=( iALURES_SIGN ||  iALURES_ZERO)?1'b1:1'b0;
		default:	  iFLAG_E=1'b0;
		endcase
	       end
	3'b001: begin
		case(SUBOP_ER) // synopsys parallel_case
		`H_SUBOP_SFEQ:  iFLAG_E=(		     iALURES_ZERO)?1'b1:1'b0;
		`H_SUBOP_SFNE:  iFLAG_E=(                 ~iALURES_ZERO)?1'b1:1'b0;
		`H_SUBOP_SFGTS: iFLAG_E=(~iALURES_SIGN && ~iALURES_ZERO)?1'b1:1'b0;
		`H_SUBOP_SFGES: iFLAG_E=(~iALURES_SIGN                 )?1'b1:1'b0;
		`H_SUBOP_SFLTS: iFLAG_E=( iALURES_SIGN && ~iALURES_ZERO)?1'b1:1'b0;
		`H_SUBOP_SFLES: iFLAG_E=( iALURES_SIGN ||  iALURES_ZERO)?1'b1:1'b0;
		default:	  iFLAG_E=1'b0;
		endcase
	       end
	default: iFLAG_E = 1'b0;	
	endcase
end

//* shift operation
	always @(AOP or BOP or SHLA_ER or SHRA_ER or SHRL_ER or ROR_ER or ROL_ER) begin
		case({SHLA_ER,SHRA_ER,SHRL_ER,ROR_ER,ROL_ER}) // synopsys parallel_case
		5'b10000:  iSH_RES       = AOP << BOP[4:0];
		5'b01000:  iSH_RES       = {{32{AOP[31]}},AOP} >> BOP[4:0];
		5'b00100:  iSH_RES       = AOP >> BOP[4:0];
		5'b00010:  iSH_RES       = {AOP,AOP} >> BOP[4:0];
		5'b00001: {iSH_RES, tmp} = {AOP,AOP} << BOP[4:0];
		default:   iSH_RES 	 = 32'hxxxx_xxxx;
		endcase
	end

//* add/sub operation
	always @(AOP or BOP or CARRY_MR or
		 SFU_ER or SFS_ER or NEGS_ER or
		 ADDU_ER or SUBU_ER or ADDS_ER or SUBS_ER or ADDWCS_ER or SUBWCS_ER) begin

		tmp1bit = 1'b0;

case({NEGS_ER,SFU_ER,SFS_ER,ADDU_ER,SUBU_ER,ADDS_ER,SUBS_ER,ADDWCS_ER,SUBWCS_ER}) // synopsys parallel_case
		9'b100000000: {iCARRY_E, iADD_RES} = ~{1'b0, BOP} + 1;
		9'b000100000: {iCARRY_E, iADD_RES} = {1'b0, AOP} + {1'b0, BOP};
		9'b010000000: begin
				iCARRY_E = 1'bx;
				{tmp1bit,  iADD_RES} = {1'b0, AOP} - {1'b0, BOP};
			      end
		9'b000010000: {iCARRY_E, iADD_RES} = {1'b0, AOP} - {1'b0, BOP};
		9'b000001000: {iCARRY_E, iADD_RES} = {AOP[31],AOP} + {BOP[31],BOP};
		9'b001000000: begin
				iCARRY_E = 1'bx;
				{tmp1bit,  iADD_RES} = {AOP[31],AOP} - {BOP[31],BOP};
			      end
		9'b000000100: {iCARRY_E, iADD_RES} = {AOP[31],AOP} - {BOP[31],BOP};
		9'b000000010: {iCARRY_E, iADD_RES} = {AOP[31],AOP} + {BOP[31],BOP} + CARRY_MR;
		9'b000000001: {iCARRY_E, iADD_RES} = {AOP[31],AOP} - {BOP[31],BOP} - CARRY_MR;
		default:      {iCARRY_E, iADD_RES} = {1'bx, 32'hxxxx_xxxx};
		endcase

		if(ADDS_ER || SUBS_ER || ADDWCS_ER || SUBWCS_ER)
			iOVERFL_E = (iCARRY_E != iADD_RES[31]) ? 1'b1 : 1'b0;
		else
			iOVERFL_E = 1'b0;
		
	end

//* logical operation
	always @(AOP or BOP or XOR_ER or OR_ER or NOTU_ER or AND_ER or CLR_ER or SET_ER) begin
		case({NOTU_ER, XOR_ER, OR_ER, AND_ER, CLR_ER, SET_ER}) // synopsys parallel_case
		6'b100000:iLOG_RES = ~BOP;
		6'b010000:iLOG_RES = AOP ^ BOP;
		6'b001000:iLOG_RES = AOP | BOP;
		6'b000100:iLOG_RES = AOP & BOP;
		6'b000001: case(BOP[4:0])
			  5'd0: iLOG_RES = AOP | 32'h0000_0001;
			  5'd1: iLOG_RES = AOP | 32'h0000_0002;
			  5'd2: iLOG_RES = AOP | 32'h0000_0004;
			  5'd3: iLOG_RES = AOP | 32'h0000_0008;
			  5'd4: iLOG_RES = AOP | 32'h0000_0010;
			  5'd5: iLOG_RES = AOP | 32'h0000_0020;
			  5'd6: iLOG_RES = AOP | 32'h0000_0040;
			  5'd7: iLOG_RES = AOP | 32'h0000_0080;
			  5'd8: iLOG_RES = AOP | 32'h0000_0100;
			  5'd9: iLOG_RES = AOP | 32'h0000_0200;
			  5'd10:iLOG_RES = AOP | 32'h0000_0400;
			  5'd11:iLOG_RES = AOP | 32'h0000_0800;
			  5'd12:iLOG_RES = AOP | 32'h0000_1000;
			  5'd13:iLOG_RES = AOP | 32'h0000_2000;
			  5'd14:iLOG_RES = AOP | 32'h0000_4000;
			  5'd15:iLOG_RES = AOP | 32'h0000_8000;
			  5'd16:iLOG_RES = AOP | 32'h0001_0000;
			  5'd17:iLOG_RES = AOP | 32'h0002_0000;
			  5'd18:iLOG_RES = AOP | 32'h0004_0000;
			  5'd19:iLOG_RES = AOP | 32'h0008_0000;
			  5'd20:iLOG_RES = AOP | 32'h0010_0000;
			  5'd21:iLOG_RES = AOP | 32'h0020_0000;
			  5'd22:iLOG_RES = AOP | 32'h0040_0000;
			  5'd23:iLOG_RES = AOP | 32'h0080_0000;
			  5'd24:iLOG_RES = AOP | 32'h0100_0000;
			  5'd25:iLOG_RES = AOP | 32'h0200_0000;
			  5'd26:iLOG_RES = AOP | 32'h0400_0000;
			  5'd27:iLOG_RES = AOP | 32'h0800_0000;
			  5'd28:iLOG_RES = AOP | 32'h1000_0000;
			  5'd29:iLOG_RES = AOP | 32'h2000_0000;
			  5'd30:iLOG_RES = AOP | 32'h4000_0000;
			  5'd31:iLOG_RES = AOP | 32'h8000_0000;
			  endcase
		6'b00010: case(BOP[4:0])
			  5'd0: iLOG_RES = AOP & 32'hffff_fffe;
			  5'd1: iLOG_RES = AOP & 32'hffff_fffd;
			  5'd2: iLOG_RES = AOP & 32'hffff_fffb;
			  5'd3: iLOG_RES = AOP & 32'hffff_fff7;
			  5'd4: iLOG_RES = AOP & 32'hffff_ffef;
			  5'd5: iLOG_RES = AOP & 32'hffff_ffdf;
			  5'd6: iLOG_RES = AOP & 32'hffff_ffbf;
			  5'd7: iLOG_RES = AOP & 32'hffff_ff7f;
			  5'd8: iLOG_RES = AOP & 32'hffff_feff;
			  5'd9: iLOG_RES = AOP & 32'hffff_fdff;
			  5'd10:iLOG_RES = AOP & 32'hffff_fbff;
			  5'd11:iLOG_RES = AOP & 32'hffff_f7ff;
			  5'd12:iLOG_RES = AOP & 32'hffff_efff;
			  5'd13:iLOG_RES = AOP & 32'hffff_dfff;
			  5'd14:iLOG_RES = AOP & 32'hffff_bfff;
			  5'd15:iLOG_RES = AOP & 32'hffff_7fff;
			  5'd16:iLOG_RES = AOP & 32'hfffe_ffff;
			  5'd17:iLOG_RES = AOP & 32'hfffd_ffff;
			  5'd18:iLOG_RES = AOP & 32'hfffb_ffff;
			  5'd19:iLOG_RES = AOP & 32'hfff7_ffff;
			  5'd20:iLOG_RES = AOP & 32'hffef_ffff;
			  5'd21:iLOG_RES = AOP & 32'hffdf_ffff;
			  5'd22:iLOG_RES = AOP & 32'hffbf_ffff;
			  5'd23:iLOG_RES = AOP & 32'hff7f_ffff;
			  5'd24:iLOG_RES = AOP & 32'hfeff_ffff;
			  5'd25:iLOG_RES = AOP & 32'hfdff_ffff;
			  5'd26:iLOG_RES = AOP & 32'hfbff_ffff;
			  5'd27:iLOG_RES = AOP & 32'hf7ff_ffff;
			  5'd28:iLOG_RES = AOP & 32'hefff_ffff;
			  5'd29:iLOG_RES = AOP & 32'hdfff_ffff;
			  5'd30:iLOG_RES = AOP & 32'hbfff_ffff;
			  5'd31:iLOG_RES = AOP & 32'h7fff_ffff;
			endcase
		default:iLOG_RES = 32'hxxxx_xxxx;
		endcase
	end

//* mov//sign/zero-extension operation
	always @(AOP or BOP or 
		 EXTHS_ER or EXTHZ_ER or EXTBS_ER or EXTBZ_ER or
		 MOV_ER or MOVIBSE_ER or MOVIBZE_ER or MOVIBLS_ER or
	  	 SWAPB_ER or SWAPH_ER
	) begin
		case({EXTHS_ER, EXTHZ_ER, EXTBS_ER, EXTBZ_ER}) // synopsys parallel_case
		4'b1000: iEXT_RES = {{16{AOP[15]}},AOP[15:0]};
		4'b0100: iEXT_RES = {16'h0000     ,AOP[15:0]};
		4'b0010: iEXT_RES = {{24{AOP[7]}} ,AOP[7:0]};
		4'b0001: iEXT_RES = {24'h000000   ,AOP[7:0]};
		default: iEXT_RES = 32'hxxxx_xxxx;
		endcase

		case({MOV_ER, MOVIBSE_ER, MOVIBZE_ER, MOVIBLS_ER, SWAPB_ER, SWAPH_ER}) // synopsys parallel_case
		6'b100000: iMOV_RES = BOP;
		6'b010000: iMOV_RES = {{24{BOP[7]}}, BOP[7:0]};
		6'b001000: iMOV_RES = {{24{1'b0  }}, BOP[7:0]};
		6'b000100: iMOV_RES = {AOP[23:0],    BOP[7:0]};
		6'b000010: iMOV_RES = {BOP[7:0], BOP[15:8], BOP[23:16], BOP[31:24]};
		6'b000001: iMOV_RES = {BOP[15:0], BOP[31:16]};
		default:   iMOV_RES = 32'hxxxx_xxxx;
		endcase
	end

	always @(iSH_RES or iADD_RES or iLOG_RES or iEXT_RES or iMOV_RES or
		 SHOP_ER or ADDOP_ER or LOGOP_ER or EXTOP_ER or MOVOP_ER
	) begin
		case({SHOP_ER, ADDOP_ER, LOGOP_ER, EXTOP_ER, MOVOP_ER}) // synopsys parallel_case
		5'b10000: iALURES = iSH_RES;
		5'b01000: iALURES = iADD_RES;
		5'b00100: iALURES = iLOG_RES;
		5'b00010: iALURES = iEXT_RES;
		5'b00001: iALURES = iMOV_RES;
		default:  iALURES = 32'hxxxx_xxxx;
		endcase
	end

endmodule
