/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: LSU.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: Load/Store Unit
//  
// Architecture Notes:
// 	Generate the address/data. 
//  
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$
`include "pool/timescale.v"
`include "pool/sys.h"
`include "pool/or1601isa.h"

module LSU(
	STALL_LSU,
	EXT_PIPE_STALL,
	iLD_DELAY_SLOT_STALL,

	iEXCEPT_M,

        iSPRG_ADDR_E,           // input from decoding stage
        BOP_ER,             // data input
        SPRG_WR_ER,

	PAIR_CNST_ER,		// PAIR auto-increment constant, it may be shifted left 2bits, so total 6 bits(signed integer)

	PAIR_MR,

	iLSU_ADDR_MR,
	LSU_ADDR,
	LSU_DO,
	LSU_DI,
        LSU_SIZE,
	LSU_BYTEN,
        LSU_LOADREQ,
        LSU_LOADREQ_MR,
        LSU_STORREQ,
        LSU_STALL,

	iLSU_DI2DP,

	AOP_ER,				// for swpai/shpai/sbpai's address
        iALURES,
	iSTOR_MR,

	LDST_PAIR_REQ_ER,
	iLDST_SZ_D,
	iLOADREQ_D,
	iLSU_SIGN_EXT_D,
	iSTORREQ_D,

	STORPAI_ER,
	STORPAI_MR,

	DADR_ALIGNEXP_MR,		// data address alignment exception
	EEAR_MR,

	CLK, 
	RSTN
);
	output		DADR_ALIGNEXP_MR;
	output [31:0]	EEAR_MR;
	input  [31:0]	AOP_ER;

	input		iEXCEPT_M;
	input		STORPAI_ER, STORPAI_MR;
	output [31:0]	PAIR_MR;
	input		LDST_PAIR_REQ_ER;
	input  [5:0]	PAIR_CNST_ER; 
	input		iLSU_SIGN_EXT_D;

	input  [31:0]	iSPRG_ADDR_E;
	input  [31:0]	BOP_ER;
	input		SPRG_WR_ER;

	output [31:0]	iLSU_DI2DP;
	input		iLD_DELAY_SLOT_STALL;

	output [31:0]	iLSU_ADDR_MR;
	output [31:0]	LSU_ADDR;
	output [31:0]	LSU_DO;
	input  [31:0]	LSU_DI;
	output [1:0]	LSU_SIZE;
	output [3:0]	LSU_BYTEN;
	output 		LSU_LOADREQ, LSU_STORREQ, LSU_LOADREQ_MR;
	output		LSU_STALL;
	input		STALL_LSU;
	input		EXT_PIPE_STALL;

	input  [31:0]	iALURES;
	input  [31:0]	iSTOR_MR;

	input  [1:0]	iLDST_SZ_D;
	input		iLOADREQ_D, iSTORREQ_D;

	input		CLK, RSTN;

//* internal signals
	reg    [1:0]	LSU_SIZE_ER, iLSU_SIZE_ER;
	reg		LSU_LOADREQ_ER, iLSU_LOADREQ_ER;
	reg		LSU_STORREQ_ER, iLSU_STORREQ_ER;
	reg		LSU_LOADREQ_MR, iLSU_LOADREQ_MR;
	reg		LSU_STORREQ_MR, iLSU_STORREQ_MR;
	reg    [31:0]	LSU_ADDR_MR, iLSU_ADDR_MR;
	reg    [31:0]	LSU_DO;
	reg		LSU_STALL_MR, iLSU_STALL_MR;
	reg   		LSU_STALL;
	reg    [1:0]	LSU_SIZE_MR;
	reg    [31:0]	LSU_DI2DP, iLSU_DI2DP;

	reg    [1:0]	M_SIZE_WR, iM_SIZE_M;
	reg    [1:0]	M_ADDR2_WR, iM_ADDR2_M;
	reg		DADR_ALIGNEXP_MR, iDADR_ALIGNEXP_MR;
	reg    [31:0]	PAIR_MR, iPAIR_MR;
	reg    [31:0]	EEAR_MR, iEEAR_MR;
	reg		iM_SIGN_EXT_M, M_SIGN_EXT_WR;
	reg		LSU_SIGN_EXT_ER, LSU_SIGN_EXT_MR;
	reg		LSU_LOADREQ_WR;
	reg		LSU_LOADREQ_MR_D1;

wire [1:0] 	LSU_SIZE    = LSU_SIZE_ER;
wire      	LSU_LOADREQ = LSU_LOADREQ_ER & ~iEXCEPT_M & ~iDADR_ALIGNEXP_MR;

wire      	LSU_STORREQ = LSU_STORREQ_ER & ~LSU_STALL & ~iEXCEPT_M & ~iDADR_ALIGNEXP_MR;

//wire [31:0]	LSU_ADDR    = (LSU_LOADREQ_ER || LSU_STORREQ_ER) ? iLSU_ADDR_MR : LSU_ADDR_MR;
//wire [31:0]	LSU_ADDR    = (~EXT_PIPE_STALL && (LSU_LOADREQ_ER || LSU_STORREQ_ER)) ? iLSU_ADDR_MR : LSU_ADDR_MR;

wire [31:0]	LSU_ADDR    = (LSU_LOADREQ_ER && ~EXT_PIPE_STALL || LSU_STORREQ_ER) ? iLSU_ADDR_MR : LSU_ADDR_MR;

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0) begin
		M_SIZE_WR	<= #(`UD) 2'b11;
		M_ADDR2_WR	<= #(`UD) 2'b00;

		LSU_SIZE_ER 	<= #(`UD) 2'b11;
		LSU_SIGN_EXT_ER	<= #(`UD) 1'b0;
		LSU_SIGN_EXT_MR	<= #(`UD) 1'b0;

		M_SIGN_EXT_WR   <= #(`UD) 1'b0;

		LSU_STORREQ_ER	<= #(`UD) 1'b0;

		LSU_LOADREQ_ER	<= #(`UD) 1'b0;
		LSU_LOADREQ_MR	<= #(`UD) 1'b0;
		LSU_LOADREQ_WR	<= #(`UD) 1'b0;
			
		PAIR_MR		<= #(`UD) 32'b0;
	end
	else if(~STALL_LSU) begin
		M_SIZE_WR	<= #(`UD) iM_SIZE_M;
		M_ADDR2_WR	<= #(`UD) iM_ADDR2_M;

		LSU_SIZE_ER 	<= #(`UD) iLSU_SIZE_ER;
		LSU_SIGN_EXT_ER	<= #(`UD) iLSU_SIGN_EXT_D;
		LSU_SIGN_EXT_MR	<= #(`UD) LSU_SIGN_EXT_ER;

		M_SIGN_EXT_WR  <= #(`UD) iM_SIGN_EXT_M;

		LSU_LOADREQ_ER	<= #(`UD) iLSU_LOADREQ_ER;
		LSU_LOADREQ_MR	<= #(`UD) iLSU_LOADREQ_MR;
		LSU_LOADREQ_WR	<= #(`UD) LSU_LOADREQ_MR;

		LSU_STORREQ_ER	<= #(`UD) iLSU_STORREQ_ER;

		PAIR_MR		<= #(`UD) iPAIR_MR;
	end

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0) begin
		LSU_STALL_MR	<= #(`UD) 1'b0;

		LSU_SIZE_MR 	<= #(`UD) 2'b11;
		LSU_ADDR_MR	<= #(`UD) 32'b0;

		DADR_ALIGNEXP_MR	<= #(`UD) 1'b0;
	end
	else if(~EXT_PIPE_STALL) begin
		LSU_STALL_MR	<= #(`UD) iLSU_STALL_MR;
		LSU_DI2DP	<= #(`UD) iLSU_DI2DP;

		LSU_SIZE_MR 	<= #(`UD) LSU_SIZE_ER;
		LSU_ADDR_MR	<= #(`UD) iLSU_ADDR_MR;
		
		DADR_ALIGNEXP_MR	<= #(`UD) iDADR_ALIGNEXP_MR;

		EEAR_MR		<= #(`UD) iEEAR_MR;
	end

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0)
		LSU_LOADREQ_MR_D1 <= #(`UD) 1'b0;
	else
		LSU_LOADREQ_MR_D1 <= #(`UD) LSU_LOADREQ_MR;

/*
	always @(posedge CLK)
	if(EXT_PIPE_STALL && LSU_LOADREQ_ER && LSU_LOADREQ_MR && ~LSU_LOADREQ_MR_D1)
		LSU_DI2DP     <= #(`UD) LSU_DI;
	else if(~EXT_PIPE_STALL)
		LSU_DI2DP     <= #(`UD) iLSU_DI2DP;
*/

always @(LSU_STALL_MR or iLSU_STALL_MR
) begin
	LSU_STALL = iLSU_STALL_MR;
end

	always @(posedge CLK)
	if(DADR_ALIGNEXP_MR)
		$display("XXXXXX LSU_DO store address alignment error at time=%d", $time);

always @(LSU_STORREQ or LSU_LOADREQ or LSU_SIZE_ER or LSU_ADDR or iEXCEPT_M or
	 DADR_ALIGNEXP_MR or LSU_LOADREQ_ER or LSU_STORREQ_ER or LSU_STALL or
	 EEAR_MR or iLSU_ADDR_MR
) begin
        if((LSU_LOADREQ_ER || LSU_STORREQ_ER && ~LSU_STALL) && ~iEXCEPT_M) begin
                //case({LSU_SIZE_ER, LSU_ADDR[1:0]}) // synopsys parallel_case
                case({LSU_SIZE_ER, iLSU_ADDR_MR[1:0]}) // synopsys parallel_case
                4'b00_00,
                4'b00_01,
                4'b00_10,
                4'b00_11,
                4'b01_00,
                4'b01_10,
                4'b11_00: begin iDADR_ALIGNEXP_MR = 1'b0; end
                default:  begin iDADR_ALIGNEXP_MR = 1'b1; end
                endcase
	end
	else begin
		iDADR_ALIGNEXP_MR = ~iEXCEPT_M & DADR_ALIGNEXP_MR;
	end

	if(iDADR_ALIGNEXP_MR)
		iEEAR_MR = LSU_ADDR[31:0];
	else
		iEEAR_MR = EEAR_MR;
end

always @(iSTOR_MR or iLSU_STALL_MR or LSU_SIZE_ER or LSU_ADDR or
 	 LSU_DI or LSU_STORREQ or
	 LSU_LOADREQ or LSU_LOADREQ_MR or LSU_SIZE_MR or LSU_ADDR_MR or
	 LSU_DI2DP or M_SIZE_WR or M_ADDR2_WR or
	 EXT_PIPE_STALL or
	 LSU_SIGN_EXT_ER or LSU_SIGN_EXT_MR or M_SIGN_EXT_WR
) begin
	if(LSU_STORREQ)
		case({LSU_SIZE_ER, LSU_ADDR[1:0]}) // synopsys parallel_case
		4'b00_00: LSU_DO = {iSTOR_MR[7:0], LSU_DI[23:0]}; //* big endian
		4'b00_01: LSU_DO = {LSU_DI[31:24], iSTOR_MR[7:0], LSU_DI[15:0]};
		4'b00_10: LSU_DO = {LSU_DI[31:16], iSTOR_MR[7:0], LSU_DI[7:0]}; 
		4'b00_11: LSU_DO = {LSU_DI[31:8], iSTOR_MR[7:0]}; 
		4'b01_00: LSU_DO = {iSTOR_MR[15:0], LSU_DI[15:0]};
		4'b01_10: LSU_DO = {LSU_DI[31:16], iSTOR_MR[15:0]};
		4'b11_00: LSU_DO = iSTOR_MR;
		default:  begin LSU_DO = iSTOR_MR; end
		//default:  begin LSU_DO = iSTOR_MR; $display("XXX LSU_DO store address align error XXX"); end
		endcase
	else
		LSU_DO = iSTOR_MR[31:16];

	iM_SIGN_EXT_M = (LSU_LOADREQ&&~LSU_LOADREQ_MR) ? LSU_SIGN_EXT_ER : (LSU_LOADREQ_MR) ? LSU_SIGN_EXT_MR  : M_SIGN_EXT_WR;

	iM_SIZE_M     = (LSU_LOADREQ&&~LSU_LOADREQ_MR) ? LSU_SIZE_ER     : (LSU_LOADREQ_MR) ? LSU_SIZE_MR      : M_SIZE_WR;
	iM_ADDR2_M    = (LSU_LOADREQ&&~LSU_LOADREQ_MR) ? LSU_ADDR[1:0]   : (LSU_LOADREQ_MR) ? LSU_ADDR_MR[1:0] : M_ADDR2_WR;

	casex({EXT_PIPE_STALL, iM_SIZE_M, iM_ADDR2_M})
	5'b000_00: iLSU_DI2DP = {{24{iM_SIGN_EXT_M & LSU_DI[31]}}, LSU_DI[31:24]};
	5'b000_01: iLSU_DI2DP = {{24{iM_SIGN_EXT_M & LSU_DI[23]}}, LSU_DI[23:16]};
	5'b000_10: iLSU_DI2DP = {{24{iM_SIGN_EXT_M & LSU_DI[15]}}, LSU_DI[15:8]};
	5'b000_11: iLSU_DI2DP = {{24{iM_SIGN_EXT_M & LSU_DI[ 7]}}, LSU_DI[7:0]}; 
	5'b001_00: iLSU_DI2DP = {{16{iM_SIGN_EXT_M & LSU_DI[31]}}, LSU_DI[31:16]};
	5'b001_10: iLSU_DI2DP = {{15{iM_SIGN_EXT_M & LSU_DI[15]}}, LSU_DI[15:0]};
	5'b011_00: iLSU_DI2DP = LSU_DI;
	5'b1xx_xx: iLSU_DI2DP = LSU_DI2DP;
	default:   iLSU_DI2DP = LSU_DI2DP;
	endcase
		
end

always @(STALL_LSU or LSU_SIZE_ER or iLOADREQ_D or iSTORREQ_D or 
	 iLDST_SZ_D or LSU_LOADREQ_ER or iLOADREQ_D or 
	 LSU_STORREQ_ER or iSTORREQ_D or 
	 iALURES or LSU_LOADREQ_MR or
	 LDST_PAIR_REQ_ER or PAIR_MR or 
	 iEXCEPT_M or STORPAI_ER
) begin
	if(iLOADREQ_D || iSTORREQ_D)
		iLSU_SIZE_ER = iLDST_SZ_D;
	else
		iLSU_SIZE_ER = 2'b11;

	iLSU_LOADREQ_ER = iLOADREQ_D;

	iLSU_LOADREQ_MR = LSU_LOADREQ_ER & ~iEXCEPT_M;

	iLSU_STORREQ_ER = iSTORREQ_D;

	if(LDST_PAIR_REQ_ER) // choose PAIR or iALURES 
		iLSU_ADDR_MR	= PAIR_MR;
	else if(STORPAI_ER)
		iLSU_ADDR_MR 	= AOP_ER;
	else
		iLSU_ADDR_MR	= iALURES;
end

// auto increment PAIR
always @(SPRG_WR_ER or iSPRG_ADDR_E or BOP_ER or 
	 LDST_PAIR_REQ_ER or PAIR_MR or PAIR_CNST_ER or
	 iDADR_ALIGNEXP_MR
) begin
	// mtsr has higher priority, since ER is written at E and auto-increment 
	// is updated at stage equivalent to M stage

	if(SPRG_WR_ER && iSPRG_ADDR_E[15:12] == `SPRG_SYS && iSPRG_ADDR_E[11:0] == `SPRG_SYS_PAIR)
		iPAIR_MR = BOP_ER;

	// not address alignment exception, one stage before iEXCEPT_M

	else if(LDST_PAIR_REQ_ER && ~iDADR_ALIGNEXP_MR) 

		iPAIR_MR = PAIR_MR + {{26{PAIR_CNST_ER[5]}}, PAIR_CNST_ER};
	else
		iPAIR_MR = PAIR_MR;
end

always @(STALL_LSU or LSU_STALL_MR or LSU_SIZE_ER or LSU_STORREQ_ER or
	 iEXCEPT_M
) begin
	if(LSU_STALL_MR)
		iLSU_STALL_MR = 1'b0;
	else if(~iEXCEPT_M && LSU_STORREQ_ER && (LSU_SIZE_ER == 2'b00 || LSU_SIZE_ER == 2'b01))
		iLSU_STALL_MR = 1'b1;
	else
		iLSU_STALL_MR = LSU_STALL_MR;
end

endmodule
