/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: PCIFA.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: Program Counter and Instruction Fetch Address generation.
//  
// Architecture Notes:
// 	It generates program counter and instruction fetch address. 
//  
//  
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$

`include "pool/timescale.v"
`include "pool/sys.h"
`include "pool/or1601isa.h"

module PCIFA(
	STALL_PCIFA, 
	EXT_PIPE_STALL,

	iALL_UPDATE_IFA_W,
	iBIU2INU_BUSY,
	EPCR_WR,          	// exception program counter

	//////////////////////////
	DADR_ALIGNEXP_MR, 	// data address alignment exception(LSU)
	II_EXP_MR, 		// illegal insn exception(DEC)
	EXT_INT_MR, 		// external interrupt(DEC)
	SYS_MR, 		// h.sys(DEC)
	BRK_MR, 		// h.brk(DEC)
	OVERFL_MR, 		// overflow exception(IALU to SPRG)
	DIVZERO_MR,		// divided by zero exception(MAC), not yet implemented
	//////////////////////////

	iJMP_D,			// Jump/Branch in Decoding stage
	iJAL_D,
	iJALR_JR_D,
	iBF_BNF_D,
	iJAL_JALR_D,

	AOP_ER,
	iAOP_ER,

	INU_DATA,
	IR_DR,
	INU_REQ,

	iIFA_W,
	PC,

	iLD_DELAY_SLOT_STALL,

	PC_PLUS_4_MR,

	iMISS_PREDICT_D, 
	iMISS_PREDICT_E, 

	iEXCEPT_M,

	CLK, 
	RSTN_D2,
	RSTN_D3
);
	output		iALL_UPDATE_IFA_W;
	input		iBIU2INU_BUSY;
	input		EXT_PIPE_STALL;
	output [31:0]	EPCR_WR;

	output		iEXCEPT_M;
	input		DADR_ALIGNEXP_MR, II_EXP_MR;
	input  [2:0]	EXT_INT_MR;
	input		SYS_MR, BRK_MR, OVERFL_MR, DIVZERO_MR;

	input		iLD_DELAY_SLOT_STALL;
	input  [31:0]	iAOP_ER, AOP_ER;

	input		iJMP_D;
	input 		iJAL_D;
	input		iJALR_JR_D, iJAL_JALR_D;
	input		iBF_BNF_D;
	input  [31:0]	INU_DATA;
	output [31:0]	IR_DR;
	output		INU_REQ;

	output [31:0]	iIFA_W;
	output [31:0]	PC, PC_PLUS_4_MR;
	input		STALL_PCIFA;
	input		iMISS_PREDICT_D, iMISS_PREDICT_E;
	input		CLK, RSTN_D2, RSTN_D3;

	reg    [31:0]	IFA_WR, iIFA_W, PC, iPC;
	reg		INU_REQ, iINU_REQ, INU_REQ_D1, INU_REQ_D2, INU_REQ_D3;
	reg    [31:0]	IR_DR, iIR_DR; 
	reg    [31:0]	iJBTA, JBTA_ER;
	reg		J_TAKEN_ER;
	reg		BF_BNF_ER;
	reg		iUPDATE_IFA_W;
	reg    [31:0]	PC_PLUS_4_ER, iPC_PLUS_4_ER;
	reg    [31:0]	PC_PLUS_4_MR, iPC_PLUS_4_MR;
	reg		iEXCEPT_M, EXCEPT_WR;
	reg    [31:0]	EPCR_MR, iEPCR_MR;
	reg    [31:0]	EPCR_WR, iEPCR_WR;
	reg    [31:0]	PC_ER, PC_MR;
	reg		iALL_UPDATE_IFA_W;

	always @(posedge CLK or negedge RSTN_D3)
	if(RSTN_D3 == 1'b0) begin
		IFA_WR		<= #(`UD) `RST_VEC;
		PC		<= #(`UD) `RST_VEC;

		IR_DR		<= #(`UD) {`H_NOP, `H_NOP};

		INU_REQ		<= #(`UD) 1'b0;
		INU_REQ_D1	<= #(`UD) 1'b0;
		INU_REQ_D2	<= #(`UD) 1'b0;

		J_TAKEN_ER	<= #(`UD) 1'b0;
	end
	else begin

		IFA_WR		<= #(`UD) iIFA_W;
		PC		<= #(`UD) iPC;

		EPCR_WR		<= #(`UD) (STALL_PCIFA) ? EPCR_WR : iEPCR_WR;
		EXCEPT_WR	<= #(`UD) (STALL_PCIFA) ? EXCEPT_WR : iEXCEPT_M;

		IR_DR		<= #(`UD) iIR_DR;

		INU_REQ		<= #(`UD) (STALL_PCIFA) ? INU_REQ : RSTN_D3;
		INU_REQ_D1	<= #(`UD) (STALL_PCIFA) ? INU_REQ_D1 : INU_REQ;
		INU_REQ_D2	<= #(`UD) (STALL_PCIFA) ? INU_REQ_D2 : INU_REQ_D1;
		INU_REQ_D3	<= #(`UD) (STALL_PCIFA) ? INU_REQ_D3 : INU_REQ_D2;

		J_TAKEN_ER	<= #(`UD) (STALL_PCIFA) ? J_TAKEN_ER : (iJMP_D | iJAL_D | iJALR_JR_D);
		BF_BNF_ER	<= #(`UD) (STALL_PCIFA) ? BF_BNF_ER : iBF_BNF_D;
		JBTA_ER		<= #(`UD) (STALL_PCIFA) ? JBTA_ER : iJBTA;

		PC_PLUS_4_ER	<= #(`UD) (STALL_PCIFA) ? PC_PLUS_4_ER : iPC_PLUS_4_ER; // return address
		PC_PLUS_4_MR	<= #(`UD) (STALL_PCIFA) ? PC_PLUS_4_MR : iPC_PLUS_4_MR; // return address

		PC_ER		<= #(`UD) (STALL_PCIFA) ? PC_ER : PC;
		PC_MR		<= #(`UD) (STALL_PCIFA) ? PC_MR : PC_ER;
	end

// calculate branch target address
always @(iJMP_D or iJAL_D or iJALR_JR_D or iBF_BNF_D or
         PC or IR_DR or
         iAOP_ER
) begin

        case({iJMP_D, iJAL_D, iBF_BNF_D, iJALR_JR_D}) // synopsys parallel_case
        4'b1000,
        4'b0100,
        4'b0010: iJBTA = PC + {{19{IR_DR[27]}}, IR_DR[27:16], 1'b0};
        4'b0001: iJBTA = iAOP_ER;
        default: iJBTA = PC;
        endcase
end

/////////////////////////////////////////////////////
// IR_DR : Instruction Register in decoding stage
/////////////////////////////////////////////////////
always @(STALL_PCIFA or INU_REQ or INU_REQ_D1 or PC or
	 IR_DR or INU_DATA or IFA_WR  or
	 J_TAKEN_ER or BF_BNF_ER or iMISS_PREDICT_E or
	 JBTA_ER or iEXCEPT_M or EXCEPT_WR or INU_REQ_D2 or INU_REQ_D3
) begin

	if(EXCEPT_WR || J_TAKEN_ER || (BF_BNF_ER & ~iMISS_PREDICT_E) ||
	   (INU_REQ_D2 && ~INU_REQ_D3) || (INU_REQ_D3 && PC[1]))
		iUPDATE_IFA_W = 1'b1;
	else
		iUPDATE_IFA_W = 1'b0;

	if(STALL_PCIFA || ~INU_REQ_D2)
		iIR_DR = IR_DR;
	else if(iEXCEPT_M)
		iIR_DR = {`H_NOP, `H_NOP};
	else if(EXCEPT_WR) 
		iIR_DR = INU_DATA;

	else if(J_TAKEN_ER || (BF_BNF_ER & ~iMISS_PREDICT_E)) 
		iIR_DR = JBTA_ER[1] ? {INU_DATA[15:0], 16'hxxxx} : INU_DATA;
	
	else if(INU_REQ_D3 && ~PC[1])
		iIR_DR = {IR_DR[15:0], INU_DATA[31:16]};
	else 
		iIR_DR = INU_DATA;
end

always @(DADR_ALIGNEXP_MR or II_EXP_MR or SYS_MR or BRK_MR or
	 OVERFL_MR or DIVZERO_MR or EXT_INT_MR
) begin
	iEXCEPT_M = DADR_ALIGNEXP_MR | II_EXP_MR | SYS_MR | BRK_MR |
	 	    OVERFL_MR | DIVZERO_MR | 
		    EXT_INT_MR[2] | EXT_INT_MR[1] | EXT_INT_MR[0];
end

`ifdef TRANSLATE_ON
	always @(posedge iEXCEPT_M) begin
	if(DADR_ALIGNEXP_MR) $display("data address alignment exception");
	if(II_EXP_MR       ) $display("illegal instruction exception");
	if(SYS_MR          ) $display("h.sys exception");
	if(BRK_MR          ) $display("h.brk exception");
	if(OVERFL_MR       ) $display("overflow exception");
	if(DIVZERO_MR      ) $display("divided by zero exception");
	if(EXT_INT_MR[2]   ) $display("EXTernal 2 INT exception");
	if(EXT_INT_MR[1]   ) $display("EXTernal 1 INT exception");
	if(EXT_INT_MR[0]   ) $display("EXTernal 0 INT exception");
	end
`endif

always @(STALL_PCIFA or IFA_WR or iJBTA or
	 iJMP_D or iJAL_D or iJALR_JR_D or iBF_BNF_D or
	 PC or
	 J_TAKEN_ER or JBTA_ER or 
	 INU_REQ or INU_REQ_D1 or 
	 iMISS_PREDICT_D or iMISS_PREDICT_E or PC or
	 J_TAKEN_ER or BF_BNF_ER or INU_REQ_D2 or INU_REQ_D3 or
	 iUPDATE_IFA_W or AOP_ER or iAOP_ER or
	 iEXCEPT_M or DADR_ALIGNEXP_MR or II_EXP_MR or
	 SYS_MR or BRK_MR or OVERFL_MR or DIVZERO_MR or
	 EXT_INT_MR or EXCEPT_WR 
) begin

/////////////////////////////////////////////////////
// iIFA_W: insn fetch address
/////////////////////////////////////////////////////
	if(iEXCEPT_M || iJALR_JR_D || iJMP_D || iJAL_D || iJALR_JR_D || 
	   iBF_BNF_D && ~iMISS_PREDICT_D || iUPDATE_IFA_W)
		iALL_UPDATE_IFA_W = 1'b1;
	else
		iALL_UPDATE_IFA_W = 1'b0;

	if(STALL_PCIFA || ~INU_REQ_D2) 
		iIFA_W = IFA_WR;
	
	else if(iEXCEPT_M) begin
	/*  
	    *** exception is prioritized ***
	    address alignment exception -> 0x600
	    illegal insn -> 0x700
	    fast external interrupt -> 0x800,0x900,0xA00
	    h.sys -> 0xC00
	    h.brk -> 0xD00
	    overfl -> 0xE00
	    divided by ZERO -> 0xF00
	*/
		casex({DADR_ALIGNEXP_MR, II_EXP_MR, EXT_INT_MR, SYS_MR, BRK_MR, OVERFL_MR, DIVZERO_MR}) // synopsys parallel_case
		9'b1xxxxxxxx: iIFA_W = 32'h0000_0600;
		9'b01xxxxxxx: iIFA_W = 32'h0000_0700;
		9'b001xxxxxx: iIFA_W = 32'h0000_0800;
		9'b0001xxxxx: iIFA_W = 32'h0000_0900;
		9'b00001xxxx: iIFA_W = 32'h0000_0A00;

		9'b000001xxx: iIFA_W = 32'h0000_0C00;
		9'b0000001xx: iIFA_W = 32'h0000_0D00;
		9'b00000001x: iIFA_W = 32'h0000_0E00;
		9'b000000001: iIFA_W = 32'h0000_0F00;
		default:      iIFA_W = 32'hxxxx_xxxx;
		endcase
	end	
	else if(iJALR_JR_D) 
		iIFA_W = iAOP_ER;
	
	else if(iJMP_D || iJAL_D || iJALR_JR_D || iBF_BNF_D && ~iMISS_PREDICT_D) 
		iIFA_W = {iJBTA[31:2], 2'b0};
	
	else if(iUPDATE_IFA_W) 
		iIFA_W = IFA_WR + 4;
	
	else 
		iIFA_W = IFA_WR;
	

/////////////////////////////////////////////////////
// PC: program counter
/////////////////////////////////////////////////////
	if(STALL_PCIFA)
		iPC = PC;
	else begin
	  if(iEXCEPT_M)
		iPC = iIFA_W;
	  else if(EXCEPT_WR)
		iPC = PC;
	  else if(J_TAKEN_ER || BF_BNF_ER && ~iMISS_PREDICT_E)
	  	iPC = JBTA_ER;	
	  else if(INU_REQ_D3)
		iPC = PC + 2;
	  else
		iPC = PC;
	end
end

always @(iEXCEPT_M or PC_MR or EPCR_WR
) begin
	/*
	DADR_ALIGNEXP_MR,       // data address alignment exception(LSU)
        II_EXP_MR,              // illegal insn exception(DEC)
        EXT_INT_MR,             // external interrupt(DEC)
        SYS_MR,                 // h.sys(DEC)
        BRK_MR,                 // h.brk(DEC)
        OVERFL_MR,              // overflow exception(IALU to SPRG)
        DIVZERO_MR
	*/

	if(iEXCEPT_M)
		iEPCR_WR = PC_MR;
	else
		iEPCR_WR = EPCR_WR;
end

always @(iJAL_JALR_D or PC or PC_PLUS_4_ER
) begin
	iPC_PLUS_4_ER = (iJAL_JALR_D) ? PC + 4 : PC_PLUS_4_ER;
	iPC_PLUS_4_MR = PC_PLUS_4_ER;
end

endmodule
