/////////////////////////////////////////////////////////////////////////////
// This file is a part of the OpenRISC 1601 Verilog-HDL model
// Copyright (C) 2000 Jimmy C. Chen and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
////////////////////////////////////////////////////////////////////////////
// IP core: OpenRISC 1601, a 16-bit instruction set/32-bit datapath RISC 
//          processor core of OR1K family, according to OR16 ISA.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * yet optimized for size (without MMU and Caches)
// 
// Documentation: http://www.opencores.org/cores/or1k/or1601/
////////////////////////////////////////////////////////////////////////////
// File: SPRG.v
// Author: Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw, 06/18/2000
// Description: SPecial Register Group module
//  
// Architecture Notes:
//  
//  
//  
// TO DO:
//
////////////////////////////////////////////////////////////////////////////
// Revision History
// $Log$
`include "pool/timescale.v"
`include "pool/sys.h"
`include "pool/or1601isa.h"

module SPRG(
	STALL_SPRG0,
	EXT_PIPE_STALL,
	iMFSR_DELAY_SLOT_STALL,

	INT_ENABLE_MR,
	EXT_INT_MR,
	EXCEPT_BJDS_WR,		// exception in branch/jump delay slot
	
	iSPRG_RD_D,
	iSPRG_WR_D,
	AOP_ER,
	BOP_ER,
	iSPRG_ADDR_E,		// input from decoding stage
	SPRG_RD_ER,		// read enable at E stage
	SPRG_RD_MR,		// read enable at E stage
	SPRG_WR_ER,		// write enable
	SPR_RES_M,		// data output

	PAIR_MR,		// PAIR_MR + increment_const
	EEAR_MR,

	RFE_MR,
	TEST_ER,
	SFU_ER,
	SFS_ER,
	iFLAG_E,
	FLAG_MR,
	FLAG_WR,

	iEXCEPT_M,
	iOVERFL_E,
	OVERFL_MR,
	OVERFL_WR,

	NEGS_ER,		// the 7 *_ER signals are used to update CARRY_MR
	ADDU_ER,
	SUBU_ER,
	ADDS_ER,
	SUBS_ER,
	ADDWCS_ER,
	SUBWCS_ER,

	iCARRY_E,
	CARRY_MR,
	CARRY_WR,

	MACHI,
	MACLO,

	EPCR_WR,		// exception program counter

	CLK,
	RSTN
);
	input		iMFSR_DELAY_SLOT_STALL;
	output		INT_ENABLE_MR;
	input  [2:0]	EXT_INT_MR;
	input		RFE_MR;
	input  [31:0]	EPCR_WR;
	input		EXCEPT_BJDS_WR;
	

	input		NEGS_ER,ADDU_ER,SUBU_ER,ADDS_ER,SUBS_ER,ADDWCS_ER,SUBWCS_ER;

	input		iSPRG_RD_D, iSPRG_WR_D;
	input  [31:0]	AOP_ER, BOP_ER;
	output [31:0]	iSPRG_ADDR_E;
	input		SPRG_RD_ER, SPRG_RD_MR, SPRG_WR_ER;
	output [31:0]	SPR_RES_M;

	input  [31:0]	PAIR_MR;
	input  [31:0]	EEAR_MR;

	input		TEST_ER, SFU_ER, SFS_ER;

	input		iFLAG_E;
	output		FLAG_MR, FLAG_WR;
	input		iEXCEPT_M;
	input		iOVERFL_E, iCARRY_E;

	output		OVERFL_MR, OVERFL_WR;
	output		CARRY_MR, CARRY_WR;
	output [31:0]	MACHI;
	output [31:0]	MACLO;
	input		STALL_SPRG0;
	input		EXT_PIPE_STALL;
	input		CLK, RSTN;

	reg    [31:0]	MACHI, MACLO, SPR_RES_M;
	reg    [31:0]	iSPRG_ADDR_E;
	reg    [31:0]	SPRG_ADDR_MR;

	reg		FLAG_MR, iFLAG_MR;
	reg		FLAG_WR, iFLAG_WR;
	reg		OVERFL_MR, iOVERFL_MR;
	reg		OVERFL_WR, iOVERFL_WR;
	reg		CARRY_MR, iCARRY_MR;
	reg		CARRY_WR, iCARRY_WR;
	reg		SUPV_WR,  iSUPV_WR;
	reg		INT_ENABLE_MR, iINT_ENABLE_MR;
	reg    [31:0]	SPR_RES_WR;

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0) begin
		CARRY_MR		<= #(`UD) 1'b0;
		CARRY_WR		<= #(`UD) 1'b0;
		FLAG_MR			<= #(`UD) 1'b0;
		FLAG_WR			<= #(`UD) 1'b0;
		SUPV_WR			<= #(`UD) 1'b0;
		INT_ENABLE_MR		<= #(`UD) 1'b0;
	end
	else begin
		CARRY_MR		<= #(`UD) (STALL_SPRG0) ? CARRY_MR : iCARRY_MR;
		CARRY_WR		<= #(`UD) (STALL_SPRG0) ? CARRY_WR : iCARRY_WR;
		FLAG_MR			<= #(`UD) (STALL_SPRG0) ? FLAG_MR  : iFLAG_MR;
		FLAG_WR			<= #(`UD) (STALL_SPRG0) ? FLAG_WR  : iFLAG_WR;

		SUPV_WR			<= #(`UD) (STALL_SPRG0) ? SUPV_WR  : iSUPV_WR;
		//SPRG_RD_MR		<= #(`UD) (STALL_SPRG0) ? SPRG_RD_MR : SPRG_RD_ER;
		INT_ENABLE_MR		<= #(`UD) (STALL_SPRG0) ? INT_ENABLE_MR : iINT_ENABLE_MR;

		SPRG_ADDR_MR		<= #(`UD) (STALL_SPRG0) ? SPRG_ADDR_MR : iSPRG_ADDR_E;
	end

	always @(posedge CLK or negedge RSTN)
	if(RSTN == 1'b0) begin
		OVERFL_MR		<= #(`UD) 1'b0;
		OVERFL_WR		<= #(`UD) 1'b0;
	end
	else if(~EXT_PIPE_STALL) begin
		OVERFL_MR		<= #(`UD) iOVERFL_MR;
		OVERFL_WR		<= #(`UD) iOVERFL_WR;

		SPR_RES_WR		<= #(`UD) SPR_RES_M;
		//SPRG_ADDR_MR		<= #(`UD) (iMFSR_DELAY_SLOT_STALL) ? SPRG_ADDR_MR : iSPRG_ADDR_E;
		//SPRG_ADDR_MR		<= #(`UD) iSPRG_ADDR_E;
	end
	
/*
	CCR(0x0000): condition code register
	31:3 	: reserved
	2	: OVERFL
	1	: CARRY
	0	: FLAG

	PAIR(0x0001) : post address increment register
	
	MACHI(0x5010): MAC hi register
	MACLO(0x5011): MAC lo register
	
 */

always @(SPRG_RD_ER or BOP_ER or AOP_ER or SPRG_WR_ER
) begin
/*
	if(SPRG_RD_ER) 
		iSPRG_ADDR_E = BOP_ER[31:0];	// mfsr
	else
		iSPRG_ADDR_E = AOP_ER[31:0]; 	// mtsr
*/
	if(SPRG_WR_ER)
                iSPRG_ADDR_E = AOP_ER[31:0];    // mtsr
        else
                iSPRG_ADDR_E = BOP_ER[31:0];    // mfsr

end

always @(SPRG_RD_ER or SPRG_RD_MR or iSPRG_ADDR_E or OVERFL_MR or EXT_INT_MR or
	 CARRY_MR or FLAG_MR or
	 MACHI or MACLO or PAIR_MR or EPCR_WR or SUPV_WR or
	 SPRG_ADDR_MR or INT_ENABLE_MR or iEXCEPT_M or
	 BOP_ER or SPRG_WR_ER or SPR_RES_WR or EXCEPT_BJDS_WR or
	 iMFSR_DELAY_SLOT_STALL or EEAR_MR
) begin

// the following part could be optimized

	if(SPRG_RD_MR && ~(SPRG_RD_MR && SPRG_RD_ER && ~iMFSR_DELAY_SLOT_STALL))
		case(SPRG_ADDR_MR[27:24]) // synopsys parallel_case
		`SPRG_SYS: case(SPRG_ADDR_MR[11:0]) // synopsys parallel_case
			   `SPRG_SYS_CCR:  SPR_RES_M = {27'b0, EXCEPT_BJDS_WR, INT_ENABLE_MR, OVERFL_MR, CARRY_MR, FLAG_MR};
			   `SPRG_SYS_MPR:  SPR_RES_M = 32'b0;
			   `SPRG_SYS_SR:   SPR_RES_M = {31'b0, SUPV_WR};
			   `SPRG_SYS_PAIR: SPR_RES_M = PAIR_MR; // if mfsr immediately follows lwpain,etc
			   `SPRG_SYS_EPCR: SPR_RES_M = EPCR_WR;	// exception PC register
			   `SPRG_SYS_EEAR: SPR_RES_M = EEAR_MR;	// exception unaligned data addr
			   default: SPR_RES_M = 32'hxxxx_xxxx;
			   endcase
		`SPRG_MAC: case(SPRG_ADDR_MR[11:0]) // synopsys parallel_case
			   `SPRG_MAC_MACHI: SPR_RES_M = MACHI;
			   `SPRG_MAC_MACLO: SPR_RES_M = MACLO;
			   default: SPR_RES_M = 32'hxxxx_xxxx;
			   endcase
		default:begin SPR_RES_M = 32'hxxxx_xxxx; $display("read unknown group register at SPRG0 at time=%d", $time); end
		endcase
	else if(SPRG_RD_ER || SPRG_RD_MR && SPRG_RD_ER && ~iMFSR_DELAY_SLOT_STALL)
		case(iSPRG_ADDR_E[27:24]) // synopsys parallel_case
		`SPRG_SYS: case(iSPRG_ADDR_E[11:0]) // synopsys parallel_case
			   `SPRG_SYS_CCR:  SPR_RES_M = {27'b0, EXCEPT_BJDS_WR, INT_ENABLE_MR, OVERFL_MR, CARRY_MR, FLAG_MR};
			   `SPRG_SYS_MPR:  SPR_RES_M = 32'b0;
			   `SPRG_SYS_SR:   SPR_RES_M = {31'b0, SUPV_WR};
			   `SPRG_SYS_PAIR: SPR_RES_M = PAIR_MR; // if mfsr immediately follows lwpain,etc
			   `SPRG_SYS_EPCR: SPR_RES_M = EPCR_WR;	// exception PC register
			   `SPRG_SYS_EEAR: SPR_RES_M = EEAR_MR;	// exception unaligned data addr
			   default: SPR_RES_M = 32'hxxxx_xxxx;
			   endcase
		`SPRG_MAC: case(iSPRG_ADDR_E[11:0]) // synopsys parallel_case
			   `SPRG_MAC_MACHI: SPR_RES_M = MACHI;
			   `SPRG_MAC_MACLO: SPR_RES_M = MACLO;
			   default: SPR_RES_M = 32'hxxxx_xxxx;
			   endcase
		default:begin SPR_RES_M = 32'hxxxx_xxxx; $display("read unknown group register at SPRG0 at time=%d", $time); end
		endcase
	else
		SPR_RES_M = 32'hxxxx_xxxx;

	if(|EXT_INT_MR)
		iINT_ENABLE_MR = 1'b0;
	else if(SPRG_WR_ER && ~iEXCEPT_M) begin
	  	if(iSPRG_ADDR_E[27:24] == `SPRG_SYS && iSPRG_ADDR_E[11:0] == `SPRG_SYS_CCR)
	          iINT_ENABLE_MR = BOP_ER[3];
	  	else
	    	  iINT_ENABLE_MR = INT_ENABLE_MR;
	end
	else
	  iINT_ENABLE_MR = INT_ENABLE_MR;
end
	
always @(iCARRY_E or CARRY_MR or iOVERFL_E or OVERFL_MR or
	 iFLAG_E or FLAG_MR or SFU_ER or SFS_ER or TEST_ER or
	 NEGS_ER or ADDU_ER or SUBU_ER or ADDS_ER or SUBS_ER or ADDWCS_ER or SUBWCS_ER or
	 iEXCEPT_M or RFE_MR or SUPV_WR or
	 SPRG_WR_ER or BOP_ER or iSPRG_ADDR_E
) begin
/* NEGS_ER,ADDU_ER,SUBU_ER,ADDS_ER,SUBS_ER,ADDWCS_ER,SUBWCS_ER */

	if(iEXCEPT_M)
		iSUPV_WR = 1'b1;
	else if(RFE_MR)
		iSUPV_WR = 1'b0;
	else
		iSUPV_WR = SUPV_WR;

	if(~iEXCEPT_M && (NEGS_ER || ADDU_ER || SUBU_ER || ADDS_ER || SUBS_ER || ADDWCS_ER || SUBWCS_ER))
		iCARRY_MR = iCARRY_E;
	else if(SPRG_WR_ER && ~iEXCEPT_M) begin
	  	if(iSPRG_ADDR_E[27:24] == `SPRG_SYS && iSPRG_ADDR_E[11:0] == `SPRG_SYS_CCR)
	          iCARRY_MR = BOP_ER[1];
		else
		  iCARRY_MR = CARRY_MR;
	end
	else
		iCARRY_MR = CARRY_MR;
	
	iCARRY_WR  = CARRY_MR;

	//if(~iEXCEPT_M && (NEGS_ER || ADDS_ER || SUBS_ER || ADDWCS_ER || SUBWCS_ER))
	if(NEGS_ER || ADDS_ER || SUBS_ER || ADDWCS_ER || SUBWCS_ER)
		iOVERFL_MR = iOVERFL_E & ~iEXCEPT_M;
	else
		iOVERFL_MR = OVERFL_MR;

	iOVERFL_WR = OVERFL_MR;

	if(~iEXCEPT_M && (SFU_ER || SFS_ER || TEST_ER))
		iFLAG_MR = iFLAG_E;
	else if(SPRG_WR_ER && ~iEXCEPT_M) begin
	  	if(iSPRG_ADDR_E[27:24] == `SPRG_SYS && iSPRG_ADDR_E[11:0] == `SPRG_SYS_CCR)
	          iFLAG_MR = BOP_ER[0];
		else
	          iFLAG_MR = FLAG_MR;
	end
	else
		iFLAG_MR = FLAG_MR;

	iFLAG_WR   = FLAG_MR;
end

always @(iCARRY_E
) begin
	MACHI 		= 32'hxxxx_xxxx;
	MACLO 		= 32'hxxxx_xxxx;
end

endmodule
